
#=====================================================================#
# procedures to create windows, input fields: char, number, lookup etc
#=====================================================================#

# Initialise window/gui setup variables, fonts and tk options
proc initialise_widgets { } {
	global widgets_initialised
	global setup

	if {[info exists widgets_initialised]} { return }
	set widgets_initialised 1

	# Delay drawing of windows until input_window is called; also defaulted in event.tcl
	setup delayed_draw true

	# Default colours
	setup select_colour "lightsteelblue"		;# The colour of the highlight bar on selection screens
	setup alert_colour "lightgoldenrod"			;# The colour of alert messages (eg: customer alerts, product alerts etc)
	setup alert_edit "lightgoldenrodYellow"		;# The colour of the alert window when mouse over for editing
	setup output_bg "lightyellow"				;# The background colour of the output windows (reports and stationery)

	# Default font used on all screens, usually set in setup.dat and varies depending on screen resolution
	setup font {Arial 12}
	setup text_font {"Courier New" 9}

	# Create bold font
	# Shouldn't need catch ...
	catch { option add *font $setup(font) startupFile }
	set setup(font_bold) [linsert $setup(font) end bold]

	# Set disabled foreground & background colour (required when using tcl >= 8.4)
	# One effect is that this sets the font colour on disabled menu items to be black; 
	# you have to configure each individual menu widget if you want anything else
	# catch errors (ie: running from tclsh rather than wish)
	# Shouldn't need catch ...
	catch {
		option add *DisabledForeground black
		option add *DisabledBackground lightgray
	}

	# Create small font (1 or 2 sizes smaller)
	# Also in printing init function ?
	set size [lindex $setup(font) 1]
	set size [expr {$size >= 10 ? 9 : $size - 1}]
	set setup(font_small) [lreplace $setup(font) 1 1 $size]
	set setup(font_small_bold) [lreplace $setup(font) 1 1 $size bold]


	# Widget scaling for high-resolution monitors
	set pixelScale [setup pixel_scale 1]

	set ::narrowBorder [expr {round(2 * $pixelScale)}]
	set ::slenderBorder [expr {round(3 * $pixelScale)}]
	set ::sleekBorder [expr {round(4 * $pixelScale)}]

	set ::skinnyPadding [expr {round(2 * $pixelScale)}]
	set ::slenderPadding [expr {round(3 * $pixelScale)}]
	set ::sleekPadding [expr {round(4 * $pixelScale)}]
	set ::narrowPadding [expr {round(5 * $pixelScale)}]

	set ::hungryPadding [expr {round(6 * $pixelScale)}]
	set ::peckishPadding [expr {round(8 * $pixelScale)}]
	set ::mediumPadding [expr {round(10 * $pixelScale)}]
	set ::plumpPadding [expr {round(12 * $pixelScale)}]

	set ::widePadding [expr {round(15 * $pixelScale)}]
}

# see input_char, set_no_trace for explanation of this bizarre proc
proc NULL { args } {}

# this procedure is used to create a new window
# wname = name of an array variable for storing the window information
# title = window title
# args	= options for the window command - to be defined...

proc create_window { wname pname title args } {
	global setup unique system

	# get the calling namespace and set the window in the global system array
	initialise_widgets
	set ns [uplevel {namespace current}]

	if { [array names system $ns,$pname] != {} } {
		if { !$::event::limbo } { event::activate_trapper $system($ns,$pname) }
	}

	# check that the window name doesn't already exist - do this to trap unwanted windows being opened
	# currently this behaviour is causing all kinds of errors.
	# usually, there is no window actually open, it just didn't get closed properly
	# Let's try to close the window that's open instead
	if { [info exists system($ns,$wname)]} {
		uplevel [list close_window $wname]
		log "The system is attempting to create a window with a name that already exists, reference \"$ns\::$wname\"."
	}

	# new window name (will be either a frame or toplevel)
	set w .window[incr unique]

	# get & set the window options
	system_set_defaults $w bg background {popup 0} position [list escape $wname] {draw_now 0}
	foreach {option value} [system_set_options $w $args] {
		error "Invalid option: $option, value = $value"
	}

	# store the window toplevel in the system array
	set system($ns,$wname) $w

	# some additional options that require setting
	set system($w,ns) $ns
	set system($w,unique) 0
	set system($w,redraw_command) ""
	set system($w,focus) ""
	set system($w,pname) $pname
	set system($w,centre) 0
	set system($w,drawn) 0
	set system($w,parent) $w

	if { !$setup(delayed_draw) } {
		set system($w,draw_now) 1
	}

	set system($w,draw_commands) {}
	set command [list draw_window $w $title $wname]
	if { $system($w,draw_now) } {
		eval $command
	} else {
		lappend system($w,draw_commands) [list $command $w]
	}

	# draw dummy 'window' to trap user-generated events
	# create_event_trapper $w

	return $w
}

proc draw_window {w title wname} {
	global setup system
	set ns $system($w,ns)

	# record the original title for instrumentation
	set original_title $title

	# create the new toplevel (set and hide/withdraw it for now) or frame
	# only -position & -popup windows create new toplevels
	if {$system($w,popup) || $system($w,position) != ""} {
		set toplevel 1
		set destroy "destroy_window $w $toplevel"

		toplevel $w
		wm title $w $title
		wm transient $w .

		# for transient windows we need to do a "grab" to stop the window behind getting any events
		set current_grab [grab current]

		# have to wait till window is visible on linux operating systems
		while { [catch { grab $w }] } {
			logToFile "grab error"
			tkwait visibility $w
		}

		if {$current_grab != ""} {
			::event::bind $w <Destroy> "+grab_original $w %W [lindex $current_grab 0]"
		}

		# The grab on the popup window causes a bug with 8.6 an greater on Windows Vista and greater
		# Specifically, if Fastbase gets minimized, the only way to get it back is by bringing
		# up the task manager and selecting the Fastbase app, right click, "switch to"
		# can be avoided by releasing the grab when we get unmapped, and then grabbing again
		# when we get remapped
		::event::bind $w <Unmap> "unmap_popup $w %W"
		::event::bind $w <Map> "map_popup $w"

		# If the main window is minimised when the popup window is drawn, then both windows will now be
		# 'trapped' on the task bar. The map binding above will re-grab when the window is deiconified/unminimised
		set state [wm state .]
		if { $state == "iconic"} {
			grab release $w
		}

		# if the current window is also a transient then add a focus command to the destroy script
		set current [lindex $setup(frame_stack) end]
		set currentf [lindex $current 0]
		if {$currentf != ""} {
			# get the focus of the current window
			# add a destroy command to reset the focus to the same field when it is re-displayed
			if {[catch { set current_focus [focus -displayof $currentf] }]} {
				set current_focus ""
			}
			append destroy ";[list focus_force $current_focus]"
		}

		# create the main bindings for the control-function keys (enquiries)
		# note: we catch errors, this is for special exception when running remote update (see fbclient)
		# because we won't be logged into a company yet, ie: no as_module() records
		catch {login::create_toplevel_bindings $w}
	} else {
		set toplevel 0
		set destroy ""

		frame $w

		# append company name to the window title
		if {$wname != "MAIN"} {
			if {[array names login::company NAME] != ""} {
				append title " - $login::company(NAME)"
			}
		}

		# Display the logged-in user's id in the window title
		if { [setup window_user_display 0] && [info exists login::user_name] && $login::user_name != ""} {
			set title "$login::user_name: $title"
		}
		
		wm title . $title

		# pack forget the current open window
		while {1} {
			set current [lindex $setup(frame_stack) end]
			set currentf [lindex $current 0]

			if {$currentf == ""} { break }
			if { ![winfo exists $currentf] } {
				# remove last invalid entry from the stack
				set setup(frame_stack) [lreplace $setup(frame_stack) end end]
				continue
			}

			# if the current window is a toplevel then withdraw it (simulate a "close")
			if {[winfo toplevel $currentf] == $currentf} {
				wm withdraw $currentf
				set setup(frame_stack) [lreplace $setup(frame_stack) end end]
				grab .

				set destroy "[list reshow_transient $currentf $current];$destroy"
				continue
			}

			# not a top-level
			break
		}

		if {$currentf != ""} {
			# get the focus of the current window
			# add a destroy command to reset the focus to the same field when it is re-displayed
			if {[catch { set current_focus [focus -displayof $currentf] }]} {
				set current_focus ""
			}
			set destroy "destroy_window $w $toplevel;$destroy;[list focus_force $current_focus]"

			pack forget $currentf

			# remove the menu from the window
			. configure -menu ""
		}

		pack $w -expand yes -fill both
	}

	# add the "window" to the frame stack
	lappend setup(frame_stack) [list $w $title $original_title]

	::event::bind $w <Destroy> +$destroy

	# -bg or -background option?
	if {$system($w,bg) != ""} { set system($w,background) $system($w,bg) }
	if {$system($w,background) != ""} {
		$w configure -bg $system($w,background)
		option add *[string trim $w .].*Background $system($w,background)
	}

	# -position?
	if {$system($w,position) != ""} {
		# position window as specified, "+x+y" notation (see "wm geometry" command)
		wm geometry $w $system($w,position)
	}

	# -popup?
	if {$system($w,popup)} {
		# withdraw "popup" window
		wm withdraw $w
		# specify that the window is to be centered later (see input_window)
		set system($w,centre) 1
	}

	# create binding so that after any event we "update_window"
	set toplevel [winfo toplevel $w]
	::event::bind $toplevel <KeyRelease> "update_window %W"
	::event::bind $toplevel <ButtonRelease> "update_window %W"

	::event::bind $w <Destroy> "+array unset system %W*;array unset system_calc %W*;array unset system_noentry %W*;array unset system_bindings %W*"

	# remove window from the system array
	::event::bind $w <Destroy> "+destroy_window_event %W $w $ns $wname"

	# -escape "window"?
	if {$system($w,escape) != ""} {
		new_bind $w <Escape> "namespace eval $ns \{ close_window $system($w,escape) destroy \}"
	}

	# trap events
	event::create_trapper $w

	set system($w,drawn) 1
}

# check if a popup window is grabbed
# if not then grab it
proc map_popup { w } {
	global setup

	# check the window is the current window on 'top' of everything else
	while 1 {
		set current [lindex $setup(frame_stack) end]
		set current_top [lindex $current 0]

		if {$current_top == ""} return
		if { ![winfo exists $current_top] } {
			# remove last invalid entry from the stack
			set setup(frame_stack) [lreplace $setup(frame_stack) end end]
			continue
		}

		break
	}
	if {$w != $current_top} return

	# check if window has grab already
	set current_grab [grab current]
	if { $current_grab == $w } return

	# finally check the window is visible
	set state [wm state $w]
	if { $state != "withdrawn" } {
		grab $w
	}
}

# release the grab on the popup wundow if the popup window becomes unmapped
proc unmap_popup { popupPath widgetPath } {
	# check the widget reporting the unmap event is the popup itself, and not a child widget
	# this happens e.g when pack forget is invoked on a child frame
	if {$popupPath eq $widgetPath} {
		grab release $popupPath
	}
}

proc focus_force { w } {
	catch { focus -force $w }
}

proc destroy_window_event { W w ns wname } {
	global system

	# only unset the system array if the destroy event is for the parent window
	if {$W == $w} {
		array unset system $ns,$wname
	}
}

proc grab_original { w1 w2 grabw } {
	if {$w1 != $w2} { return }
	catch { grab $grabw }
}

proc reshow_transient { w current } {
	global setup

	lappend setup(frame_stack) $current

	# re-display the window and set the grab
	# check that the window exists first, as we had some errors occuring at Briteuro, not sure why yet
	# see also: login/delete_main_window
	if {[winfo exists $w]} {
		wm deiconify $w
		grab $w
	}
}

proc destroy_window { w toplevel } {
	global setup

	set data [lindex $setup(frame_stack) end]
	set win [lindex $data 0]

	# check the window being destroyed is the current top "window"
	if {$win != $w} {
		# is this an error or is the destroy event getting called multiple times?
		return
	}

	# remove the last entry from the frame stack
	set setup(frame_stack) [lreplace $setup(frame_stack) end end]

	if {$toplevel} { return }

	# check if any more windows open?
	if {![llength $setup(frame_stack)]} {
		destroy .
		return
	}

	set data [lindex $setup(frame_stack) end]
	set w [lindex $data 0]
	set title [lindex $data 1]

	wm title . $title

	# check if the menu is to be re-displayed
	if {[llength $setup(frame_stack)] == 1} {
		. configure -menu .menu
	}

	# re-display the previous "window"
	# catch errors (not sure why, but Carey from corpcons has got errors here, to do with zeacom/popups)
	catch {
		pack $w -expand yes -fill both
	}
}

proc new_bind { window event command } {
	global system_bindings

	set toplevel [winfo toplevel $window]
	set topw [top_window $window]

	# determine if this command is already specified for this window/binding, if it is then don't add it again
	# an example of a duplicate binding occurs when changing company (the F1,F2... bindings for enquiries)
	set add_command 1
	if {![catch { set commands $system_bindings($window,$event) }]} {
		# get list of commands
		if {[lsearch $commands $command] != -1} {
			set add_command 0
		}
	}

	if {$add_command} {
		lappend system_bindings($window,$event) $command
	}

	# get the current list of bindings for this window/event, add "new_bind_event" if not there already
	set bindings [bind $toplevel $event]
	if {![string match "*new_bind_event*" $bindings]} {
		::event::bind $toplevel $event "+new_bind_event %W $event"
	}
}

proc new_bind_event { window event } {
	global setup system_bindings

	# get the current window (top screen)
	set currentf [lindex [lindex $setup(frame_stack) end] 0]

	# if the current window matches the binding then execute the command(s)
	foreach name [lsort [array names system_bindings *,$event]] {
		set w [lindex [split $name ","] 0]
		set topw [top_window $w]
		if {$topw == $currentf} {
			# detect errors, after evaluating cmd system bindings may get deleted from under us (eg: if command closes the current window)
			if {![catch { set bindings $system_bindings($name) }]} {
				foreach cmd  $bindings {
					# evaluate each command
					eval $cmd
				}
			}
		}
	}
}

#====================================================================#

# centre a window in the middle of the screen
# will be on the primary monitor if multi-monitor setup
proc centre_window { w } {
	if {[info commands $w] == ""} { return }

	set xmax [winfo screenwidth $w]
	set ymax [winfo screenheight $w]
	set x [expr {($xmax - [winfo reqwidth $w]) / 2}]
	set y [expr {($ymax - [winfo reqheight $w]) / 2}]

	wm geometry $w "+$x+$y"
}

# centre popup window in the middle of the main fastbase screen
# check window does not go over edge of screen
proc centre_popup_window { w } {
	if {[info commands $w] == ""} { return }

	set current_geometry [winfo geometry $w]
	scan $current_geometry "%dx%d+%d+%d" width height x y

	# find the dimensions of the root toplevel
	# [wm geometry .] does not return the correct geometry of a maximized window
	# Instead, it returns the width & height of the maximized window, but the x & y position
	# of the window before it was maximized. [winfo geometry .] returns the position and
	# size of the maximized window
	# set parent_geometry [wm geometry .]
	set parent_geometry [winfo geometry .]
	scan $parent_geometry "%dx%d+%d+%d" pwidth pheight px py

	set newx [expr {$px + ($pwidth-[winfo reqwidth $w])/2}]
	set newy [expr {$py + ($pheight-[winfo reqheight $w])/2}]

	set left_edge 	[winfo vrootx $w]
	set top_edge 	[winfo vrooty $w]
	set right_edge 	[winfo vrootwidth $w]
	set bottom_edge [winfo vrootheight $w]

	# horizontal position first: does the window cross the right hand side
	if {$newx+$width>$right_edge} {
		# move the right edge of the window left to the right edge of the screen
		set newx [expr {$right_edge-$width}]
	}
	# window is off screen to the left
	if {$newx<$left_edge} {
		set newx $left_edge
	}
	# if the window is too wide for the screen, the overflow will be on the right

	# vertical position: same logic as above
	if {$newy+$height>$bottom_edge} {
		set newy [expr {$bottom_edge-$height}]
	}
	if {$newy<$top_edge} {
		set newy $top_edge
	}

	wm geometry $w "+$newx+$newy"
}

#====================================================================#

# display a list of proxy hosts to the user.
# return the selected proxy host
# if user closes the window, reurns an empty string
proc select_proxy_host { proxy_host_list } {

	namespace eval ::proxySelection {
		set hostList {}
		set databaseList {}
		set selectedHost ""
		set selectedDatabase ""
		set confirmed 0

		proc set_proxy_host { proxyListWidget databaseListWidget} {
			variable selectedHost
			variable selectedDatabase
			variable hostList
			variable databaseList
			
			set value [$proxyListWidget get]

			# display button on screen for all servers
			foreach hostInfo $hostList {
				set name [lindex $hostInfo 0]
				if {[string match $name $value]} {
					set selectedHost [lrange $hostInfo 0 end-1]
					set databaseList [lindex $hostInfo end]
					set selectedDatabase ""

					# reset the database list
					$databaseListWidget list delete 0 end
					if {[$databaseListWidget cget -state] != "normal"} {
						$databaseListWidget configure -state normal -background SystemWindow
					}
					# add the databases for this host to the list
					foreach databaseName $databaseList {
						$databaseListWidget list insert end $databaseName
					}
					# set initial selection to first option
					$databaseListWidget configure -value [lindex $databaseList 0]
					pack .selectProxy.df.l2 $databaseListWidget -pady $::narrowPadding -padx $::hungryPadding -side left

					if { [llength $databaseList] <= 1 } {
						$databaseListWidget configure -state "disabled" -background lightgray
					}

					return
				}
			}
		}

		proc confirm_proxy_host { databaseListWidget } {
			variable selectedDatabase
			variable databaseList
			variable confirmed

			set value [$databaseListWidget get]
			foreach databaseName $databaseList {
				if { $datbaseName == $value } {
				set selectedDatabase $databaseName
					break
				}
			}
			set confirmed 1
		}
	}

	set ::proxySelection::hostList $proxy_host_list

	wm withdraw .
	wm title . "Fastbase"
	frame .selectProxy -borderwidth 3 -relief groove
	pack .selectProxy

	set pixelScale [setup pixel_scale 1]
	set widgetWidth [expr {round(35 * $pixelScale)}]

	label .selectProxy.l2 -text "Fastbase" -font {Arial 14}
	pack .selectProxy.l2 -padx $::hungryPadding

	pack [frame .selectProxy.pf]
	label .selectProxy.pf.l2 -text "Select Server" -font {Arial 14}
	set proxyListWidget [combobox::combobox .selectProxy.pf.ff -width $widgetWidth -editable 0 -takefocus 1 -font {Arial 14} -bg gray95]

	# display button on screen for all servers
	foreach server $proxy_host_list {
		set name [lindex $server 0]
		$proxyListWidget list insert end $name
	}

	# set initial selection to first option
	$proxyListWidget configure -value [lindex [lindex $proxy_host_list 0] 0]

	pack [frame .selectProxy.df]
	label .selectProxy.df.l2 -text "Select Database" -font {Arial 14}
	set databaseListWidget [combobox::combobox .selectProxy.ff -width $widgetWidth -editable 0 -takefocus 1 -font {Arial 14} -bg gray95]

	set b [button .selectProxy.fb -text "Connect" -command "confirm_proxy_host $proxyListWidget" -font {Arial 14 bold} -fg darkred]

	pack .selectProxy.pf.l2 $proxyListWidget -pady $::narrowPadding -padx $::hungryPadding -side left
	pack .selectProxy.df.l2 $databaseListWidget -pady $::narrowPadding -padx $::hungryPadding -side left
	pack $b

	# when a host is selected, update/default the database
	$proxyListWidget configure -command "::proxySelection::set_proxy_host $proxyListWidget $databaseListWidget"

	::proxySelection::set_proxy_host $proxyListWidget $databaseListWidget

	# escape key will close window
	::event::bind . <Escape> "set proxySelection::selectedHost \"\"; set proxySelection::confirmed \"\""

	# centre window
	wm resizable . 0 0
	update
	set xmax [winfo screenwidth .]
	set ymax [winfo screenheight .]
	set x [expr {($xmax - [winfo reqwidth .]) / 2}]
	set y [expr {($ymax - [winfo reqheight .]) / 2}]
	wm geometry . "+$x+$y"

	focus -force $proxyListWidget
	tkwait var proxySelection::confirmed

	# remove the selection frame
	destroy .selectProxy
	wm resizable . 1 1

	# destory the binding created above
	::event::bind . <Escape> ""

	return [list $::proxySelection::selectedHost $::proxySelection::selectedDatabase]
}

#====================================================================#

# this procedure is used to create a new notebook within a window
# fName = name of an array variable for storing the frame information
# wname = name of the parent window array
# title = window title
# args	= options for the window command - to be defined

# the command results in extra arrays being created for each tab
# each array represents a new window, original = lw, tabs = lw1, lw2 etc.

proc create_notebook { wname args } {
	# get the namespace of the calling procedure, this will be used when
	# referencing all widget variables, stored in "wname(ns)"
	global system
	set ns [uplevel {namespace current}]

	# define a window name
	set f [system_new_window $wname "tab"]

	# define all options
	system_set_defaults $f tabs {width 550} {height 300} {id ""}
	set unused [system_set_options $f $args]

	set pages 0
	set page_list {}

	foreach page $system($f,tabs) {
		lappend page_list $page

		set tn $f.body.f[incr pages]
		set vname $wname$pages

		set system($tn,unique) 0
		set system($ns,$vname) $tn
		system_parent_window $tn $ns $wname
	}

	set command "[list draw_notebook $f $page_list] $unused"
	system_input_widget $f $ns $command $wname

	return $f
}

proc draw_notebook { f pages args} {
	global system
	global notebooks

	# if this notebook has an id, find the last-raised tab for this id
	# before creating the notebook, as the Rnotebook:create command will reset this back to 1
	set initialTab ""
	if {$system($f,id) ne ""} {
		set id $system($f,id)
		if {[info exists notebooks($id)]} {
			set initialTab $notebooks($id)
		} else {
			# firstime this notebook has been drawn, default page 1
			set notebooks($id) 1
		}
	}

	Rnotebook:create $f -tabs $pages -width $system($f,width) -height $system($f,height) -onchangetab [list notebook_tab_selected $f]
	pack $f -padx 0 -pady 0 -expand yes -fill both

	# process all options except -tabs now that widget is created
	foreach { option value } $args {
		$f configure $option $value
	}

	# raise the initial tab, if there is one
	# can be enabled/disabled via setup option remember_tabs
	if {$initialTab ne "" && [setup remember_tabs 0]} { Rnotebook:raise $f $initialTab }

	# add keyboard bindings for next/previous tab pages? control-left/right
	new_bind $f <Control-Left> "Rnotebook:raise $f -1"
	new_bind $f <Control-Right> "Rnotebook:raise $f +1"

	set system($f,drawn) 1
}

# record the last selected tab for the notebook
# so that when another notebook of the same id is drawn, we can return to the same tab
proc notebook_tab_selected { f tab } {
	global system
	global notebooks

	if {![info exists system($f,id)]} return

	if {$system($f,id) ne ""} {
		set id $system($f,id)
		set notebooks($id) $tab
	}
}

#====================================================================#

# this procedure is used to create a new frame within a window
# fName = name of an array variable for storing the frame information
# wname = name of the parent window array
# title = window title
# args	= options for the window command - to be defined

proc create_frame { fname wname title args } {
	global system setup
	set ns [uplevel {namespace current}]

	# define a window name
	set f [system_new_window $wname]

	# define all options
	system_set_defaults $f {pad 0} {side "top"} {anchor "nw"} background bg {expand "no"} {border 0} {fill "both"} {padx ""} {progress 0}

	# store options and process undefined ones
	set draw_args [system_set_options $f $args]

	# save the standard options required for windows/frames
	set system($f,unique) 0
	set system($f,ns) $ns
	# get the calling namespace and set the window in the global system array
	set system($ns,$fname) [list $f $f]

	set command "[list draw_frame $f $fname $title] $draw_args"
	system_input_widget $f $ns $command $wname

	return $f
}

proc draw_canvas { f fname args } {
	global system setup

	set ns $system($f,ns)
		canvas $f 
		set fc $f

	# process undefined options
	foreach {option value} $args {
		$f configure $option $value
	}

	# if -padx not set then use -pad setting
	if {$system($f,padx) == ""} { set system($f,padx) $system($f,pad) }

	pack $f -padx $system($f,padx) -pady $system($f,pad) \
		-side $system($f,side) -anchor $system($f,anchor) \
		-expand $system($f,expand) -fill $system($f,fill)


	# -bg or -background option?
	if {$system($f,bg) != ""} { set system($f,background) $system($f,bg) }
	if {$system($f,background) != ""} {
		$f configure -bg $system($f,background)
		option add *[string trim $f .].*Background $system($f,background)
	}

	# -border? this option may be used if there is no frame title to still draw a groove border
	if {$system($f,border)} {
		$f configure -relief groove -borderwidth 2
	}

	set system($f,drawn) 1
}

proc draw_frame { f fname title args } {
	global system setup

	set ns $system($f,ns)

	# either use our special labelled frame widget or the standard frame?
	if {$title != ""} {
		# test for win98, if so remove the text as it stuffs things up
		if {[setup labelframe_bug 0]} {
			global tcl_platform
			if {$tcl_platform(os) == "Windows 95"} { set title "" }
		}

		labelframe $f -text $title -font [setup font_small_bold] -relief groove -highlightthickness 2 -highlightcolor gray -padx $::skinnyPadding
		set fc $f
	} else {
		frame $f -bd 0
		set fc $f
	}

	# process undefined options
	foreach {option value} $args {
		$f configure $option $value
	}

	# if -padx not set then use -pad setting
	if {$system($f,padx) == ""} { set system($f,padx) $system($f,pad) }

	pack $f -padx $system($f,padx) -pady $system($f,pad) \
		-side $system($f,side) -anchor $system($f,anchor) \
		-expand $system($f,expand) -fill $system($f,fill)

	# Create progress bar on frame
	if {$system($f,progress)} {
		progress::create_local $f
	}

	# -bg or -background option?
	if {$system($f,bg) != ""} { set system($f,background) $system($f,bg) }
	if {$system($f,background) != ""} {
		$f configure -bg $system($f,background)
		option add *[string trim $f .].*Background $system($f,background)
	}

	# -border? this option may be used if there is no frame title to still draw a groove border
	if {$system($f,border)} {
		$f configure -relief groove -borderwidth 2
	}

	# include binding to destroy all system variables associated with the frames components
	::event::bind $f <Destroy> "+array unset system %W*;array unset system_calc %W*;array unset system_noentry %W*;array unset system_bindings %W*"

	# remove window from the system array (same as close_window but for frames)
	::event::bind $f <Destroy> "+array unset system $ns,$fname"

	set system($f,drawn) 1
}

#====================================================================#

proc create_separator { wname } {
	global system

	set f [system_new_window $wname]
	set ns [uplevel {namespace current}]

	set command [list draw_separator $f]
	system_input_widget $f $ns $command $wname

	return $f
}

proc draw_separator { f } {
	frame $f -borderwidth 1 -relief sunken -height $::narrowBorder
	pack $f -padx $::plumpPadding -pady $::narrowPadding -fill x

	set system($f,drawn) 1
}

#====================================================================#

# wname = Fastbase name of the parent window - see "create_window", usually 'w' or similar
# field = name of the variable to store the data
# title = label text (ie: screen prompt)
# args	= options as follows:
#	-required 0/1	(force entry of field)
#	-password char	(password entry, eg: -password "\u2022" bullet char)
# 	-event {event_name command ...} generic event handling. Events are evaluated without validation
# 	-note (string value) or -notevar (variable name) for explanatory text displayed to right of field (-notevar for dynamic text)
proc input_char { wname field title args } {
	global system setup

	set ns [uplevel {namespace current}]
	set f [system_new_window $wname]
	set fe $f.e

	set system($fe,ns) $ns
	set system($fe,field) $field
	set system($fe,wname) $wname

	# define all options
	system_set_defaults $fe format size width uppercase lowercase date password standard pluskey \
		validate valid {singlequote 0} required noentry enterkey enterkeycontinue enterkeytrap event range warning_range filename help help_args \
		note notevar note_config var calc labelvar {endofmonth 0} {startofmonth 0} {email 0} {url 0} {align 1} \
		labelwidth label_config {thismonth 0} {whitespace 1} \
		measure measure_unit measure_display_unit measure_display_unit_var

	# store options and process undefined ones
	set draw_args [system_set_options $fe $args]

	set command "[list draw_input_char $f $fe $title] $draw_args"
	system_input_widget $fe $ns $command

	return $fe
}

proc draw_input_char {f fe title args} {
	global system setup

	set ns $system($fe,ns)
	set field $system($fe,field)

	set labelw [lindex [pc_label $f $title] 1]
	entry $fe -validate all -validatecommand "entry_validation char %W %V %d %S %P"
	set_focus $fe
	pack $fe -side left -anchor nw

	foreach {option value} $args {
		$fe configure $option $value
	}

	# -password? display bullet or whatever character was specified (value)
	if {$system($fe,password) != ""} {
		$fe configure -show $system($fe,password)
	}

	# -enterkey; process event when user tabs/enters out of field
	if {$system($fe,enterkey) != ""} {
		# note: the -enterkey option validates the screen before action
		# use: -event {Tab command} for non-validation
		set cmd "if \{\[validate_window $fe\]\} \{ $system($fe,enterkey);break \} else \{ break \}"
		::event::bind $fe <Tab> +[list namespace eval $ns $cmd]
	}

	# -enterkeycontinue; process event when user tabs/enters out of field, and don't stop unless validation fails
	if {$system($fe,enterkeycontinue) != ""} {
		# note: the -enterkeycontinue option validates the screen before action
		# use: -event {Tab command} for non-validation
		set cmd "if \{\[validate_window $fe\]\} \{ $system($fe,enterkeycontinue) \} else \{ break \}"
		::event::bind $fe <Tab> +[list namespace eval $ns $cmd]
	}

	# -enterkeytrap; trap any user events and process event when user tabs/enters out of field
	if {$system($fe,enterkeytrap) != ""} {
		# note: the -enterkeytrap option validates the screen before action
		# use: -event {Tab command} for non-validation
		set parent [top_window $fe]
		set cmd "if \{\[validate_window $fe\]\} \{ event::activate_trapper $parent; $system($fe,enterkeytrap); event::disable_trapper; break \} else \{ break \}"
		::event::bind $fe <Tab> +[list namespace eval $ns $cmd]
	}

	# -pluskey; as for -enterkey, but only enabled by setup option enable_plus_key
	if {$system($fe,pluskey) != "" && [setup enable_plus_key 0]} {
		# note: the -pluskey option validates the screen before action
		set cmd "if \{\[validate_window $fe\]\} \{ $system($fe,pluskey);break \} else \{ break \}"
		::event::bind $fe <KeyPress-plus> +[list namespace eval $ns $cmd]
	}
	
	# -event? generic event handling: -event {event_name command ...}
	if {$system($fe,event) != ""} {
		foreach {event command} $system($fe,event) {
			# execute the specified command when the event occurs
			::event::bind $fe <$event> "+if \{\[[list namespace eval $ns $command]\]\} break"
		}
	}

	# -width?
	if {$system($fe,width) != ""} {
		$fe configure -width $system($fe,width)
	}

	# -labelvar?
	if {$system($fe,labelvar) != ""} {
		$labelw configure -textvariable $ns\::$system($fe,labelvar)
	}

	# -label_config?
	if {$system($fe,label_config) != ""} {
		eval "$labelw configure $system($fe,label_config)"
	}

	# -align 0? this option removes the widget from the packing/alignment list (see proc pack_labels)
	if {!$system($fe,align)} {
		array unset system $labelw,pack
	}
	# -labelwidth? set the label width to the specified size, used in conjunction with "-align 0"
	if {$system($fe,labelwidth) != ""} {
		$labelw configure -width $system($fe,labelwidth)
	}

	# -date field?
	if {$system($fe,date) == "1"} {
		# display the date field (and set trace if variable changes)
		if {[info exists $ns\::$field] == 0} { set $ns\::$field "" }

		# ensure the date field is no more than 10 characters, ie: doesn't include the time 00:00:00
		set $ns\::$field [string range [set $ns\::$field] 0 9]

		$fe insert end [date::display [set $ns\::$field]]

		# entry_validation sets the variable value on key changes
		# entry_validation sets the display on focusout events
		$fe configure -validatecommand "entry_validation date %W %V %d %S %P"

		# standard validation logic checks the date validity
		set oldDate [set $ns\::$field]
		set cmd "\[date::validate \[$fe get\] $system($fe,endofmonth) $system($fe,startofmonth) $oldDate $system($fe,thismonth)\]"
		set system($fe,valiDATE) [list $cmd "Invalid Date"]

		# set write trace to update variable on screen if changed
		set cmd "NULL $fe;update_input_display date $ns\::$field $fe"
		trace variable $ns\::$field w $cmd
		::event::bind $fe <Destroy> +[list trace vdelete $ns\::$field w $cmd]

		# keypresses to alter the date (page up/down)
		::event::bind $fe <Key-Prior> "user_change_date $fe Prior"
		::event::bind $fe <Key-Next> "user_change_date $fe Next"

		# double click or press F1 on date field? select date from calendar
		::event::bind $fe <Control-F1> "+NULL"
		::event::bind $fe <KeyPress-F1> "+date::calendar_select $fe;break"
		::event::bind $fe <Double-1> "+date::calendar_select $fe;break"
	}

	# -time field?
	if {$system($fe,date) == "2"} {
		# display the time field (and set trace if variable changes)
		if {[info exists $ns\::$field] == 0} { set $ns\::$field "" }

		$fe insert end [set $ns\::$field]

		# entry_validation sets the variable value on key changes
		# entry_validation sets the display on focusout events
		$fe configure -validatecommand "entry_validation time %W %V %d %S %P"

		# standard validation logic checks the time validity
		set cmd "\[date::validate_time \[$fe get\]\]"
		set system($fe,valiDATE) [list $cmd "Invalid Time"]

		# set write trace to update variable on screen if changed
		set cmd "NULL $fe;update_input_display time $ns\::$field $fe"
		trace variable $ns\::$field w $cmd
		::event::bind $fe <Destroy> +[list trace vdelete $ns\::$field w $cmd]

		# keypresses to alter the date (page up/down)
		::event::bind $fe <Key-Prior> "user_change_time $fe Prior"
		::event::bind $fe <Key-Next> "user_change_time $fe Next"
	}

	# -measure field?
	if {$system($fe,measure)=="distance"} {
		# display the measurement (and set trace if variable changes)
		if {[info exists $ns\::$field] == 0} { set $ns\::$field "" }

		if {$system($fe,measure_display_unit_var)!=""} {
			set system($fe,measure_display_unit) [set $ns\::$system($fe,measure_display_unit_var)]
		}

		# default display unit to same as actual unit
		if {$system($fe,measure_display_unit)==""} {
			set system($fe,measure_display_unit) $system($fe,measure_unit)
		}
		# convert from current unit to display unit
		set converted_distance [measure::convert_distance [set $ns\::$field] $system($fe,measure_unit) $system($fe,measure_display_unit)]

		# give the converted value to the entry widget
		$fe insert end [measure::display_distance $converted_distance $system($fe,measure_display_unit)]

		# entry_validation sets the variable value on key changes
		# entry_validation sets the display on focusout events
		$fe configure -validatecommand "entry_validation distance %W %V %d %S %P"

		# standard validation logic checks the validity
		set cmd "\[measure::validate_distance \[$fe get\] \[set ::system($fe,measure_display_unit)\]\]"
		set system($fe,valiDATE) [list $cmd "Invalid Measurement"]

		# set write trace to update variable on screen if changed
		set cmd "NULL $fe;update_input_display distance $ns\::$field $fe"
		trace variable $ns\::$field w $cmd
		::event::bind $fe <Destroy> +[list trace vdelete $ns\::$field w $cmd]

		# display the units to the right
		set note [measure::distance_input_note [set $ns\::$field] $system($fe,measure_unit) $system($fe,measure_display_unit)]
		label $f.din -text $note -justify left -font [setup font_small]
		pack $f.din -side left
		# record the path of the note widget in the system array so it can be updated when the value changes
		set system($fe,measure_note_label) $f.din

		# if the units are tied to a variable, create a trace to update the field if it changes...
		if {$system($fe,measure_display_unit_var)!=""} {
			set cmd "NULL $fe;update_input_distance_unit $fe"
			trace variable $ns\::$system($fe,measure_display_unit_var) w $cmd
			::event::bind $fe <Destroy> +[list trace vdelete $ns\::$system($fe,measure_display_unit_var) w $cmd]
		}
	}

	# -note? note to display to right of the entry box?
	set added_note 0
	if {$system($fe,note) != ""} {
		label $f.t -text $system($fe,note) -justify left -font [setup font_small]
		pack $f.t -side left
		set added_note 1
	} elseif {$system($fe,notevar) != ""} {
		label $f.t -textvariable $ns\::$system($fe,notevar) -justify left -font [setup font_small]
		pack $f.t -side left
		set added_note 1
	}

	# -note_config?
	if {$added_note && $system($fe,note_config) != ""} {
		eval "$f.t configure $system($fe,note_config)"
	}

	# -var?
	if {$system($fe,var) != ""} {
		label $f.v -textvariable $ns\::$system($fe,var) -justify left -font [setup font_small]
		pack $f.v -side left
	}

	# -filename {open/save [options]}
	if {$system($fe,filename) != ""} {
		# parameter = open or save
		# eg: -filename {open -filetypes {{SQL .sql}}}
		# F1 on entry or click on label to activate file dialog
		set value $system($fe,filename)

		set options [lrange $value 1 end]
		command::arguments $options dialog_options {} {title "" filetypes {} initialdir "" defaultextension ""}

		set parent [winfo toplevel $fe]
		set title $dialog_options(title)
		set filetypes $dialog_options(filetypes)
		set directory $dialog_options(initialdir)
		set extension $dialog_options(defaultextension)

		switch -- [lindex $value 0] {
			"save" {
				# set script_elements [list {+set x [file_dialog::choose_writable "} $parent {" "} $title {" "} $filetypes {" "} $directory {" [set } "${ns}::${field}" {] "} $extension {"]}]
				# set cmd [join $script_elements ""]

				set cmd "+set x \[file_dialog::choose_writable \"$parent\" \"$title\" \"$filetypes\" \"$directory\" \[set ${ns}::${field}\] \"$extension\"\]; "
				# set cmd "+set x \[tk_getSaveFile -initialdir [pwd] -initialfile \[set $ns\::$field\]"
			}
			"dir" {
				set cmd "+set x \[tk_chooseDirectory -initialdir \[set $ns\::$field\]"
			}
			default {
				set cmd "+set x \[file_dialog::choose_readable \"$parent\" \"$title\" \"$filetypes\" \"$directory\" \[set ${ns}::${field}\] \"$extension\"\]; "
				# set cmd "+set x \[tk_getOpenFile -initialdir [pwd] -initialfile \[set $ns\::$field\]"
			}
		}


		set fb $f.b
		label $fb -text "  \[Browse ...\]  " -relief flat -bd $::narrowBorder -fg darkred
		::event::bind $fb <Enter> "+$fb configure -relief raised"
		::event::bind $fb <Leave> "+$fb configure -relief flat"

		pack $fb -side left -padx $::peckishPadding

		# set-up bindings to activate "file dialog" window
		append cmd "if \{\"\$x\" != \"\"\} \{set $ns\::$field \$x\}"
		::event::bind $fb <1> $cmd
		::event::bind $fe <Control-F1> "+NULL"
		::event::bind $fe <Key-F1> $cmd
	}

	# -help text?
	if {$system($fe,help) != ""} {
		set cmd [list popup_help $system($fe,help) $fe]
		# append all help arguments onto the command; help_args might be used for "-time 0" or other, see support.tcl - popup_help
		foreach arg $system($fe,help_args) { lappend cmd $arg }
		::event::bind $fe <FocusIn> $cmd
	}

	# -email?
	if {$system($fe,email)} {
		# provisional option to allow user to click on email and start an email message via "mailto"
		# bind $fe <Enter> "popup_help \"Click to Start an Email Message.\" $fe"
		# bind $fe <1> "exec [lindex [auto_execok start] 0] mailto:\[set $ns\::$field\] &"

		# add binding to F1 - search for customer, F2 - search for user, F3 - search for supplier
		set enter_script ""
		append enter_script {
			set widget } $fe {
			if {[$widget cget -state] ne "disabled"} {
				popup_help "F1 - $login::ar_module(ACCOUNT_NAME) search; F2 - user search; F3 - $login::ap_module(ACCOUNT_NAME)" $widget
			}
		}
		::event::bind $fe <Enter> +$enter_script

		::event::bind $fe <Control-F1> "+NULL"
		::event::bind $fe <Control-F2> "+NULL"
		::event::bind $fe <Control-F3> "+NULL"

		::event::bind $fe <KeyPress-F1> +[list choose_customer_email $fe]
		::event::bind $fe <Key-F2> +[list choose_user_email $fe]
		::event::bind $fe <Key-F3> +[list choose_supplier_email $fe]

		# validate the email address
		set cmd "\[validate_email \[$fe get\]\]"
		set system($fe,valiDATE) [list $cmd "Invalid Email Address"]
	}

	# link the entryfield to the variable/field (date/measure fields are special exceptions)
	if {$system($fe,date) == "" && $system($fe,measure) == ""} {
		$fe configure -textvariable $ns\::$field
	}

	# set the size/width of the field
	# if uppercase then increase the width to allow for larger characters
	if {$system($fe,size) != "" && $system($fe,width) == ""} {
		set sizex [expr {$system($fe,size) + 2}]
		if {$system($fe,uppercase) == "1"} {
			set sizex [expr {int([expr {$sizex * 1.25}])}]
		}
		$fe configure -width $sizex
	}

	# -standard "code" text selection option?
	if {$system($fe,standard) != ""} {
		# press F1 to select from list of options
		# press F2 to maintain list of options (requires level 1 UT security)
		::event::bind $fe <Control-F1> "+NULL"
		::event::bind $fe <KeyPress-F1> "+standard_text_selection $fe"
		set f2 ""
		if {$login::user(UT_SECURITY)} {
			::event::bind $fe <Control-F2> "+NULL"
			::event::bind $fe <KeyPress-F2> "+standard_text_selection $fe maint"
			set f2 "\nPress F2 to maintain messages."
		}
		::event::bind $fe <FocusIn> "popup_help \"Press F1 to select message.$f2\" $fe -xoffset 200"
	}

	# -uppercase/lowercase?
	if {$system($fe,uppercase) == "1" || $system($fe,lowercase) == "1"} {
		# set write trace to update variable (uppercase/lowercase)
		set cmd "NULL $fe;update_case [list $ns\::$field $system($fe,uppercase)]"
		trace variable $ns\::$field w $cmd
		::event::bind $fe <Destroy> +[list trace vdelete $ns\::$field w $cmd]
	}

	# arrange for a menu to appear upon right-click
	set extras {}

	# if the field isn't explicitly upper or lower case then provide an
	# option to convert to upper or lower case
	set case_override [expr {$system($fe,uppercase) != "" || $system($fe,lowercase) != ""}]
	if {!$case_override} {
		lappend extras "U" "L"
	}

	# url ? add option to visit website
	if {$system($fe,url)} { lappend extras "URL" }

	::event::bind $fe <ButtonPress-3> "application_popup_entry $fe %X %Y $extras"

	set_system $fe calc
	set_system $fe noentry

	set system($fe,drawn) 1
}

# Map internal event to tk event (with the angle brackets)
proc map_event { event } {
	if {$event ne "Shift-Tab"} { return "<${event}>" }
	return [expr {$::tcl_platform(platform) eq "unix" ? "<Shift-ISO_Left_Tab>" : "<Shift-Tab>"}]
}

# Update the displayed value in a Fastbase input widget
# Called when the variable associated with a fastbase input widget changes
# Invoked via a tcl variable write trace (see input_char)
# Only used for input_date, input_time, input_distance
# All other input_char widgets use the -textvariable option on the tk entry widget which automatically updates
#   the displayed value whenever the value changes
# type is either date, time, distance
# field is name of the field associated with the widget
# w is the path of the tk entry widget
# name1 name2 op are all added automatically by the tcl trace and are not used by this proc
# No value is returned
proc update_input_display { type field w name1 name2 op} {
	global system

	# check the tk entry widget still exists
	if {[info commands $w] != ""} {
		# save the state of the date field (-state, -validate)
		set state [$w cget -state]
		set valid [$w cget -validate]

		# make the field normal (so we can change its value) and disable validation
		$w configure -state normal -validate none
		$w delete 0 end

		set value [set $field]

		# displayed value depends on the type of field
		switch -- $type {
			"date" { set display_value [date::display $value] }
			"time" { set display_value $value }
			"distance" {
				set store_unit $system($w,measure_unit)
				set display_unit $system($w,measure_display_unit)

				# convert the value to the display unit before formatting it
				set display_value [measure::convert_distance $value $store_unit $display_unit]
				set display_value [measure::display_distance $display_value $display_unit]

				# update the display note to the right of the field
				set note [measure::distance_input_note $value $store_unit $display_unit]
				$system($w,measure_note_label) configure -text $note
			}
		}

		$w insert end $display_value

		# reset the status and validation
		$w configure -state $state -validate $valid
	}
}

proc user_change_date { fe method } {
	global system

	set ns $system($fe,ns)
	set field $system($fe,field)

	set date [set $ns\::$field]

	if {$date == ""} { return 1 }

	switch -- $method {
		"Prior" {
			set new [date::calc $date - months 1]
			if {$date == [date::endofmonth $date]} { set new [date::endofmonth $new] }
		}
		"Next" {
			set new [date::calc $date + months 1]
			if {$date == [date::endofmonth $date]} { set new [date::endofmonth $new] }
		}
	}

	# set new date
	set $ns\::$field $new
	return 1
}

proc user_change_time { fe method } {
	global system

	set ns $system($fe,ns)
	set field $system($fe,field)

	set time [set $ns\::$field]

	if {$time == ""} { return 1 }

	# get hour
	set hour [string trimleft [string range $time 0 1] "0"]
	if {$hour == ""} { set hour 0 }

	switch -- $method {
		"Prior" {
			incr hour -1
			if {$hour < 0} { set hour 23 }
		}
		"Next" {
			incr hour
			if {$hour > 23} { set hour 0 }
		}
	}

	set new [format %02d $hour]
	append new [string range $time 2 4]

	# set new date
	set $ns\::$field $new
	return 1
}

proc update_case { field upper name1 name2 op } {
	# convert the field to upper or lower case as specified
	set case [expr {$upper == "1"} ? {"toupper"} : {"tolower"}]
	set $field [string $case [set $field]]
}

proc validate_email { email } {
	# validate the email address, nothing is also valid
	if {$email == ""} { return 1 }
	return [regexp {^(.+)@(.+)$} $email]
}

# update the valriable associated with an entry widget without triggering the trace that updates the displayed value
proc set_no_trace { var value w } {
	# build a list of traces to delete
	# we only delete traces that start with the command "NULL"
	# and are for the same window as the one we are setting; 03-04-2001
	# all traces which we don't want executed will be defined with the NULL
	# command at the start to identify themselves as "delete me"
	set disabled_traces [disable_entry_variable_traces $var $w]

	# set the variable to the desired value
	set $var $value

	enable_entry_variable_traces $var $disabled_traces
}

# get the value of the valriable associated with an entry widget without triggering the trace that updates the displayed value
proc get_no_trace { var w } {
	set disabled_traces [disable_entry_variable_traces $var $w]

	# set the variable to the desired value
	set value [set $var]

	enable_entry_variable_traces $var $disabled_traces

	return $value
}

# disable any traces on the variable associated with the entry widget w
# return the list of traces disabled
proc disable_entry_variable_traces { var w } {
	# build a list of traces to delete
	# we only delete traces that start with the command "NULL"
	# and are for the same widget as the one we are setting; 03-04-2001
	# all traces which we don't want executed will be defined with the NULL
	# command at the start to identify themselves as "delete me"

	set null_string "NULL $w"
	set null_length [expr {[string length $null_string] - 1}]

	set traces {}
	foreach trace [trace vinfo $var] {
		if {[string range [lindex $trace 1] 0 $null_length] == $null_string} {
			lappend traces $trace
		}
	}

	# remove existing traces
	foreach trace $traces {
		trace vdelete $var [lindex $trace 0] [lindex $trace 1]
	}

	return $traces
}

# (re) enable traces on the variable 
proc enable_entry_variable_traces { var trace_list } {
	# reset/add back the original traces
	foreach trace $trace_list {
		trace variable $var [lindex $trace 0] [lindex $trace 1]
	}
}

#====================================================================#

proc set_focus { w {action ""} } {
	global system
	set top [top_window $w]

	# if focus not set yet then set it to the first field
	# return 1 if set
	if {$system($top,focus) == "" || $action == "force"} {
		set system($top,focus) $w
		return 1
	}

	return 0
}

proc top_window { w } {
	# return the top "window" for the specified widget
	return ".[lindex [split $w .] 1]"
}

#====================================================================#

# This procedure is used to check the value entered in an input field is valid. Used for the following Fastbase input widgets:
#    input_char, input_email, input_phone, input_url, input_number, input_date, input_distance
# It is called whenever the tk entry widget for one of these fastbase widgets registers one of the following events:
#   focusin - when the keyboard focus enters the field
#   focusout - when the keyboard focus leaves the field
#   key - when the user presses a key
# type is the type of entry field, can be either
#   number - input_number
#   date - input_date
#   distance - input_distance
#   char - all others
# w is the tk path of the entry widget
# event is the type of event reported by the tk entry widget, either focusin, focusout, key (see above)
# command is the type of edit action reported by the entry widget:
#   -1 means nothing has changed, usually a focus event
#   0 is delete, this means some or all of the value in the entry field has been deleted
#   1 is insert, this means some new text is being inserted into the entry field 
# string is the text being inserted or deleted (will be empty string if command is -1)
# new_value is what the value in the text field will be if the entry is valid (i.e if this procedure returns true)
# The return value is either true if the entry is valid, or false if the entry is invalid
# For date, time, and measure fields this procedure will return true even if the new value is invalid
#   This is because the user must be allowed to (for example) have '12/' entered in a date field
proc entry_validation { type w event command string new_value } {
	global system

	# if the field is a date, number etc (i.e requires validation at a type level) then attempt to update the field value
	if {$type != "char"} {
		set ns $system($w,ns)
		set field $system($w,field)
		set value [string trim $new_value]

		set valid 0
		switch -- $type {
			"date" {
				# check the original date is set
				if {[catch { set system($w,original_date) }]} {
					set system($w,original_date) [set $ns\::$field]
				}

				# if the entered date is valid set the value (only update display on "focusout")
				if {[date::validate $value 0 0 $system($w,original_date)]} {
					set valid 1
					set value [date::store $value $system($w,original_date)]
				}
			}
			"time" {
				# if the entered time is valid set the field (only update display on "focusout")
				if {[date::validate_time $value]} {
					set valid 1
					set value [date::store_time $value]
				}
			}
			"distance" {
				set store_unit $system($w,measure_unit)
				set display_unit $system($w,measure_display_unit)

				# Measurements are stored in mm, up to 6dp
				# The 6dp is required for imperial fraction values, which go up to 64ths
				# However, only 2dp are displayed for metric values
				# So if the measurement unit  for an input_distance field is switched from imperial to metric, 
				#    the value will actually be altered at this point if it uses more than 2dp in mm
				if {[measure::validate_distance $value $display_unit]} {
					# new distance is valid
					set valid 1

					# convert the entered value into the correct units (normalise first)
					set value [measure::normalise_distance $value $display_unit]
					set value [measure::convert_distance $value $display_unit $store_unit]

					# update the text to the right of the field
					set note [measure::distance_input_note $value $store_unit $display_unit]
					$system($w,measure_note_label) configure -text $note
				}
			}
			"number" {
				set valid 1

				if { $system($w,blankok) == 1 && $new_value == ""} {
					set value ""
				} else {
					set value [entry_number $w $new_value]
				}
			}
		}

		# update the value
		if {$valid} {
			set_no_trace $ns\::$field $value $w
		}

		if {$event == "focusout"} {
			# if the value is invalid, return true anyway
			# field will be validated later (e.g when user presses save button) and user will be shown validation error
			if {!$valid} {return 1}

			switch -- $type {
				"date" {
					# date field? expand the date field on screen if valid
					# if the date field is valid set the field & update the display
					set display_value [date::display [set $ns\::$field]]
				}
				"time" {
					# if the time field is valid set the field & update the display
					set display_value [date::store_time [set $ns\::$field]]
				}
				"distance" {
					set store_unit $system($w,measure_unit)
					set display_unit $system($w,measure_display_unit)
					set display_value [measure::convert_distance [set $ns\::$field] $store_unit $display_unit]
					set display_value [measure::display_distance $display_value $display_unit]
				}
				"number" {
					set display_value [set $ns\::$field]

				}
			}

			# update the displayed value
			$w configure -validate none
			$w delete 0 end
			$w insert end $display_value
			$w configure -validate all

			return 1
		}
	}

	# key? validate the insertion
	# other event? ok
	if {$event != "key"} {
		return 1
	}

	# only interested in "insert (1)" events
	if {$command != 1} {
		return 1
	}

	# check if there is a size restriction on the field?
	if {![catch { set size $system($w,size) }]} {
		if {$size != ""} {
			if {[string length $new_value] > $size} {
				bell
				return 0
			}
		}
	}

	# check if there is a set of valid entries?
	if {![catch { set valid $system($w,valid) }]} {
		if {$valid != ""} {
			# process each character in the inserted string (are they valid?)
			for {set i 0} {$i < [string length $string]} {incr i} {
				if {[string first [string range $string $i $i] $valid] == -1} {
					bell
					return 0
				}
			}
		}
	}

	# do we allow singlequotes to be entered, the default is no
	if {[catch { set singlequote $system($w,singlequote) } ]} {
		set singlequote 0
	}

	# we are purposely excluding the single quote character from entry as it causes SQL problems
	if {[string first "'" $new_value] != -1 && !$singlequote} {
		bell
		return 0
	}

	# do we allow whitespace to be entered, the default is yes
	if {[info exists system($w,whitespace)] && $system($w,whitespace) == 0} {
		if {[string first " " $new_value] != -1} {
			bell
			return 0
		}
	}

	return 1
}

proc entry_number { w new_value } {
	global system

	set format $system($w,format)
	set number [string trim $new_value]

	# catch any errors processing entered number/expression
	if {[catch {set number [expr {$number + 0}]}]} {
		set number 0
	}

	set number [string trimleft [format "\%$format\x66" $number]]
	return $number
}

# the measurement display unit of an input_distance widget has changed
# w is the tk entry path
# other args are added by tcl trace and not used
proc update_input_distance_unit {w name1 name2 op} {
	global system

	# check the tk entry widget still exists
	if {[info commands $w] != ""} {
		set ns $system($w,ns)
		set field $system($w,field)

		set new_display_unit [get_no_trace $ns\::$system($w,measure_display_unit_var) $w]

		# update the display unit in the system array
		set system($w,measure_display_unit) $new_display_unit
		# force the value to be redisplayed by setting the field (but not changing it)
		set $ns\::$field [set $ns\::$field]
	}
}

#====================================================================#

proc input_phone { wname field title args } {
	set cmd "input_char $wname $field \"$title\" -size 30 -width 30 $args"
	uplevel 1 $cmd
}

proc input_url { wname field title args } {
	set cmd "input_char $wname $field \"$title\" -size 100 -width 50 -url 1 $args"
	uplevel 1 $cmd
}

proc input_email { wname field title args } {
	set cmd "input_char $wname $field \"$title\" -size 100 -width 35 -email 1 -singlequote 1 -fg darkgreen -disabledforeground darkgreen $args"
	uplevel 1 $cmd
}

proc choose_customer_email { fe } {
	global system
	variable selected_email_customer

	set selected_email_customer(NUMBER) 0
	set selected_email_customer(EMAIL_ADDRESS) ""

	# display find customer widget; update the value of the field with the chosen customer's email
	set wname $system($fe,wname)
	set cw $wname\__choose_ar_email[incr unique]
	create_window $cw $wname "Find $login::ar_module(ACCOUNT_NAME) email" -popup 1

	input_account $cw selected_email_customer(NUMBER) AR $login::ar_module(ACCOUNT_NAME) -array selected_email_customer
	input_char $cw selected_email_customer(EMAIL_ADDRESS) "Email" -noentry 1 -width 70

	input_button $cw -buttons [list [list "OK" [list customer_email_chosen $fe] -image "save"]]
	input_window $cw
}

proc customer_email_chosen { fe } {
	global system
	variable selected_email_customer

	if {$selected_email_customer(NUMBER) == 0} return
	if {$selected_email_customer(EMAIL_ADDRESS) eq ""} return

	set ns $system($fe,ns)
	set field $system($fe,field)

	set ${ns}::${field} $selected_email_customer(EMAIL_ADDRESS)
}

proc choose_user_email { fe } {
	global system
	variable selected_email_user

	set selected_email_user(USER_ID) ""
	set selected_email_user(EMAIL_ADDRESS) ""

	# display find user widget; update the value of the field with the chosen user's email
	set wname $system($fe,wname)
	set uw $wname\__choose_user_email[incr unique]
	create_window $uw $wname "Find User email" -popup 1

	input_user $uw selected_email_user(USER_ID) "User" -onchange load_email_user
	input_char $uw selected_email_user(EMAIL_ADDRESS) "Email" -noentry 1 -width 50

	input_button $uw -buttons [list [list "OK" [list user_email_chosen $fe] -image "save"]]
	input_window $uw
}

proc load_email_user {} {
	variable selected_email_user
	query::find selected_email_user USER -id {USER_ID}
}

proc user_email_chosen { fe } {
	global system
	variable selected_email_user

	if {$selected_email_user(USER_ID) eq ""} return
	if {$selected_email_user(EMAIL_ADDRESS) eq ""} return

	set ns $system($fe,ns)
	set field $system($fe,field)

	set ${ns}::${field} $selected_email_user(EMAIL_ADDRESS)
}

proc choose_supplier_email { fe } {
	global system
	variable selected_email_supplier

	set selected_email_supplier(NUMBER) 0
	set selected_email_supplier(EMAIL_ADDRESS) ""

	# display find supplier widget; update the value of the field with the chosen supplier's email
	set wname $system($fe,wname)
	set sw $wname\__choose_ap_email[incr unique]
	create_window $sw $wname "Find $login::ap_module(ACCOUNT_NAME) email" -popup 1

	input_account $sw selected_email_supplier(NUMBER) AP $login::ap_module(ACCOUNT_NAME) -array selected_email_supplier
	input_char $sw selected_email_supplier(EMAIL_ADDRESS) "Email" -noentry 1 -width 70

	input_button $sw -buttons [list [list "OK" [list supplier_email_chosen $fe] -image "save"]]
	input_window $sw
}

proc supplier_email_chosen { fe } {
	global system
	variable selected_email_supplier

	if {$selected_email_supplier(NUMBER) == 0} return
	if {$selected_email_supplier(EMAIL_ADDRESS) eq ""} return

	set ns $system($fe,ns)
	set field $system($fe,field)

	set ${ns}::${field} $selected_email_supplier(EMAIL_ADDRESS)
}

# used for e.g entering branch id - no whitespace or other 'dangerous' chars (backslash, curly braces etc)
# characters like +, &, $, # excluded on somewhat arbitrary grounds I suppose -- just don't think an ID should have them
# use -number 1 for ids that are integers
proc input_id { wname field title args } {

	set valid_characters {*-./0123456789:@ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz~}
	set index [lsearch $args "-number"]
	if {$index >= 0} {
		# number id - value must be integer
		set number_id [lindex $args ${index}+1]
		set args [lreplace $args $index ${index}+1]

		if {$number_id} { set valid_characters {0123456789} }
	}

	set cmd [list input_char $wname $field $title]
	# lappend cmd -size 30 -width 30
	lappend cmd -uppercase 1
	lappend cmd -whitespace 0
	lappend cmd -valid $valid_characters
	lappend cmd {*}$args
	uplevel 1 $cmd
}

#====================================================================#

proc input_date { wname field title args } {
	set cmd "input_char $wname $field \"$title\" -date 1 -size 10 -valid {-/0123456789.} $args"
	uplevel 1 $cmd
}

proc input_time { wname field title args } {
	set cmd "input_char $wname $field \"$title\" -date 2 -size 5 -valid {0123456789:} $args"
	uplevel 1 $cmd
}

proc input_distance { wname field title args } {
	# what size.... (-size option)
	set cmd "input_char $wname $field \"$title\" -measure distance -valid {_/0123456789.} $args"
	uplevel 1 $cmd
}

#====================================================================#

# wname = name of the window array - see "create_window"
# field = name of the field to store the data
# title = field title (ie: screen prompt)
# args	= options as follows:
#	    -format 10.4	(defines field as 10 before and 4 after ".")
#	    -low value		(lower limit) accepts expression
#	    -high value 	(upper limit) accepts expression
#	    -note text		(text to display after number, eg "%"}
#	    -blankok		(value can be an empty string)
# the following notes added by nathan, they may not be 100% correct
#		-var variable_name value of variable will display after number. If the value of the variable changes, so will the text displayed  
#		-customer customer_number
#		-noentry boolean (is the field display only. 1 = Yes, 0 = No) accepts expression
#		-validate
#		-event
#		-bg
#		-help
#		-help_args

proc input_number { wname field title args } {
	global system setup
	set f [system_new_window $wname]
	set fe $f.c

	set ns [uplevel {namespace current}]
	set system($fe,ns) $ns
	set system($fe,field) $field
	set system($fe,window) $fe
	set system($fe,onchangeX) [set $ns\::$field]

	# define all options
	system_set_defaults $fe required noentry high low {format 10.2} onchange width pluskey \
		note notevar note_config var var_config validate enterkey event labelvar help help_args \
		{align 1} labelwidth label_config {blankok 0}

	# store options and process undefined ones
	set draw_args [system_set_options $fe $args]

	set command "[list draw_input_number $f $fe $title] $draw_args"
	system_input_widget $fe $ns $command $wname

	return $fe
}

proc draw_input_number { f fe title args} {
	global system setup

	set ns $system($fe,ns)

	set labelw [lindex [pc_label $f $title] 1]
	entry $fe -validate all -validatecommand "entry_validation number %W %V %d %S %P"
	set_focus $fe
	pack $fe -side left -anchor nw

	foreach {option value} $args {
		$fe configure $option $value
	}

	# -note? note to display to right of the entry box?
	if {$system($fe,note) != ""} {
		label $f.t -text $system($fe,note) -justify left -font [setup font_small]
		pack $f.t -side left
	} elseif {$system($fe,notevar) != ""} {
		label $f.t -textvariable $ns\::$system($fe,notevar) -justify left -font [setup font_small]
		pack $f.t -side left
	}

	if {$system($fe,note_config) != ""} {
		eval "$f.t configure $system($fe,note_config)"
	}

	# -var?
	if {$system($fe,var) != ""} {
		label $f.v -textvariable $ns\::$system($fe,var) -justify left -font [setup font_small]
		pack $f.v -side left

		if {$system($fe,var_config) != ""} {
			eval "$f.v configure $system($fe,var_config)"
		}
	}

	# -enterkey; process event when user tabs/enters out of field
	if {$system($fe,enterkey) != ""} {
		# note: the -enterkey option validates the screen before action
		# use: -event {Tab command} for non-validation
		set cmd "if \{\[validate_window $fe\]\} \{ $system($fe,enterkey);break \} else \{ break \}"
		::event::bind $fe <Tab> +[list namespace eval $ns $cmd]
	}
	
	# -pluskey; as for -enterkey, but only enabled by setup option enable_plus_key
	if {$system($fe,pluskey) != "" && [setup enable_plus_key 0]} {
		# note: the -pluskey option validates the screen before action
		set cmd "if \{\[validate_window $fe\]\} \{ $system($fe,pluskey);break \} else \{ break \}"
		::event::bind $fe <KeyPress-plus> +[list namespace eval $ns $cmd]
	}
	
	# -event? generic event handling: -event {event_name command ...}
	if {$system($fe,event) != ""} {
		foreach {event command} $system($fe,event) {
			# execute the specified command when the event occurs
			::event::bind $fe <$event> "+if \{\[[list namespace eval $ns $command]\]\} break"
		}
	}

	# -labelvar?
	if {$system($fe,labelvar) != ""} {
		$labelw configure -textvariable $ns\::$system($fe,labelvar)
	}

	# -label_config?
	if {$system($fe,label_config) != ""} {
		eval "$labelw configure $system($fe,label_config)"
	}

	# -format? update value to correct details for numerics (.0?)
	if {[string first "." $system($fe,format)] == -1} {
		append system($fe,format) ".0"
	}

	# -align 0? this option removes the widget from the packing/alignment list (see proc pack_labels)
	if {!$system($fe,align)} {
		array unset system $labelw,pack
	}
	# -labelwidth? set the label width to the specified size, used in conjunction with "-align 0"
	if {$system($fe,labelwidth) != ""} {
		$labelw configure -width $system($fe,labelwidth)
	}

	# -help text?
	if {$system($fe,help) != ""} {
		set cmd [list popup_help $system($fe,help) $fe]
		# append all help arguments onto the command; help_args might be used for "-time 0" or other, see support.tcl - popup_help
		foreach arg $system($fe,help_args) { lappend cmd $arg }
		::event::bind $fe <FocusIn> $cmd
	}

	# -onchange command?
	if {$system($fe,onchange) != ""} {
		set system($fe,calc) "input::check_onchange $fe"
		set_system $fe calc
	}

	input_number_window $fe

	# override the default width?
	if {$system($fe,width) != ""} {
		$fe configure -width $system($fe,width)
	}

	# allow entry of digits & basic expression stuff
	set system($fe,valid) { ()*+-/.0123456789}

	# pop-up menu accessible via the right-mouse-click
	::event::bind $fe <ButtonPress-3> "application_popup_entry $fe %X %Y"

	set_system $fe noentry

	set system($fe,drawn) 1
}

proc update_number { fe name1 name2 op } {
	global system

	set ns $system($fe,ns)
	set field $system($fe,field)
	set format $system($fe,format)

	if {[info commands $fe] != ""} {
		set number [set $ns\::$field]
		if {$number == "" && $system($fe,blankok) == 0} {
			set number 0
		}
		if { $number != "" } {
			# round the number using internal procedure first, to avoid rounding errors
			# possibly a check should be done here to make sure the number is not infinity
			# (new from Tcl 8.5, see comments above system::round proc)
			# currently just ignoring any errors...
			set dp [lindex [split $format "."] 1]
			if { $dp != "" } {
				catch { set number [system::round $number $dp] }
			}
			set number [string trimleft [format "\%$format\x66" $number]]
		}
		set state [$fe cget -state]
		set selection [$fe selection present]
		$fe configure -state normal
		$fe delete 0 end
		$fe insert end $number
		$fe configure -state $state

		# if the field was selected before then re-select entire contents
		if {$selection} {
			$fe selection range 0 end
		}
	}
}

proc input_number_window { f } {
	global system
	set format $system($f,format)
	set ns $system($f,ns)
	set field $system($f,field)
	set window $system($f,window)

	set width [system::get_num_width $format]

	$window configure -width $width -justify right

	set number [set $ns\::$field]
	if {$number == "" && $system($f,blankok) == 0} {
		set number 0
	}
	if { $number != "" } {
		# round the number using internal procedure first, to avoid rounding errors
		set dp [lindex [split $format "."] 1]
		if { $dp != "" } {
			catch { set number [system::round $number $dp] }
		}
		set number [string trimleft [format "\%$format\x66" $number]]
	}
	$window insert end $number

	# set write trace to update variable on screen if changed
	set cmd "NULL $f;update_number $f"
	trace variable $ns\::$field w $cmd
	::event::bind $window <Destroy> +[list trace vdelete $ns\::$field w $cmd]
}

#====================================================================#

# wname  = name of the window array - see "create_window"
# args	 = options as follows:
# NOTE THAT: the orient & takefocus buttons must come before the actions, buttons, or maint options! Or they won't apply!
#   -buttons {name action misc} (also list, eg: {{1 ""} {2 ""}} etc
#	    name   = button text
#	    action = command to action when button is selected
#	    misc   = "none" by default, "novalidate" = no validation (ie: active)
#	    image = image to use (override default)
#   -actions {} same as the "buttons" option except the window is not closed!
#   -maint {record new? [delete_cmd]}
#	    this is a quick way to have generic save, delete, cancel buttons
#	    record = the record name
#	    new?   = 0/1 if new record (ie: disable delete button)
#	    delete_cmd = command to execute if delete button pressed (optional)
#   -orient h/v : horizontal (default) or vertical
#   -progress 0/1 : create progress bar
# 	-takefocus 0/1 : should the button take keyboard focus?

proc input_button { wname args } {
	global system

	set b [system_new_window $wname]
	set ns [uplevel {namespace current}]

	set command "[list draw_input_button $b $wname $ns] $args"
	system_input_widget $b $ns $command $wname

	return $b
}

proc draw_input_button { b wname ns args } {
	global system setup

	frame $b
	set buttons 0
	pack $b -fill x

	set orient "left"
	set fill "none"
	set takefocus 1

	# process options
	foreach { option value } $args {
		switch -- $option {
			-orient {
				if {$value == "v"} {
					set orient top
					set fill "x"
				} else {
					set orient $value
					set fill "none"
				}
			}
			-takefocus {
				set takefocus $value
			}
			-buttons {
				foreach button $value {
					set bx $b.b[incr buttons]

					set button_args [lrange $button 2 end]
					array set button_options [process_button_args $button_args]

					# title2 is title with underscores removed
					set title [lindex $button 0]
					regsub -all "_" $title "" title2

					set cmd [lindex $button 1]

					if { $button_options(instrument) } {
						set cmd "instrument_button \"$title2\"; $cmd"
					}

					if { [setup delayed_draw 0] && $button_options(delayclose) } {
						set cmd "close_window $wname -delay;$cmd"
					} else {
						set cmd "close_window $wname;$cmd"
					}

					if { !$button_options(novalidate) } {
						set cmd "if \{\[validate_window $b\]\} \{ $cmd; progress::finish_local \}"
					}

					# determine whether we use an image button or standard?
					# if an image exists with the first word then assume image
					if {$button_options(image) == ""} {
						set name [string tolower [lindex $title2 0]]
						if {[make_image $name]} {
							set name image_$name
						} else {
							set name image_unknown
						}
					} else {
						set name image_$button_options(image)
					}

					# special "none"?
					if {$name == "image_none"} { set name "" }

					eval "[list pc_button $bx $name $title [list namespace eval $ns $cmd]] $button_options(tk_args)"
				
					if {!$takefocus} { $bx configure -takefocus 0 }
					pack $bx -side $orient -padx $::sleekPadding -pady $::skinnyPadding -anchor w -fill $fill

					if { $button_options(report_info) != "" } {
						::event::bind $bx <ButtonPress-3> "application_popup_entry $bx %X %Y REPORT_BUTTON \"$title2\" \"$button_options(report_info)\""
					} else {
						::event::bind $bx <ButtonPress-3> "application_popup_entry $bx %X %Y BUTTON \"$title2\""
					}
				}
			}
			-actions {
				foreach button $value {
					set bx $b.b[incr buttons]
					set cmd "[lindex $button 1]"

					set button_args [lrange $button 2 end]
					array set button_options [process_button_args $button_args]

					# title2 is title with underscores removed
					set title [lindex $button 0]
					regsub -all "_" $title "" title2

					if { $button_options(instrument) } {
						set cmd "instrument_button \"$title2\"; $cmd"
					}

					if { $button_options(trapevents) } {
						set parent $system($b,parent)
						set cmd "event::activate_trapper $parent; $cmd; event::disable_trapper"
					}

					if { !$button_options(novalidate) } {
						set cmd "if \{\[validate_window $b\]\} \{ $cmd; progress::finish_local \}"
					}

					# determine whether we use an image button or standard?
					# if an image exists with the first word then assume image

					if {$button_options(image) == ""} {
						set name [string tolower [lindex $title2 0]]
						if {[make_image $name]} {
							set name image_$name
						} else {
							set name image_unknown
						}
					} else {
						set name image_$button_options(image)
					}

					# special "none"?
					if {$name == "image_none"} { set name "" }
					
					eval "[list pc_button $bx $name $title [list namespace eval $ns $cmd]] $button_options(tk_args)"
					if {!$takefocus} { $bx configure -takefocus 0 }
					pack $bx -side $orient -padx $::sleekPadding -pady $::skinnyPadding -anchor w -fill $fill

					if { $button_options(report_info) != "" } {
						::event::bind $bx <ButtonPress-3> "application_popup_entry $bx %X %Y REPORT_BUTTON \"$title2\" \"$button_options(report_info)\""
					} else {
						::event::bind $bx <ButtonPress-3> "application_popup_entry $bx %X %Y BUTTON \"$title2\""
					}
				}
			}
			-help {
				set bx $b.b[incr buttons]

				set cmd "instrument_button \"Help\"; "
				append cmd "help_link \"[lindex $value 0]\""

				pc_button $bx image_help "_Help" $cmd
				if {!$takefocus} { $bx configure -takefocus 0 }

				pack $bx -side $orient -padx $::sleekPadding -pady $::skinnyPadding -anchor w -fill $fill
				::event::bind $bx <ButtonPress-3> "application_popup_entry $bx %X %Y BUTTON \"Help\""
			}
			-maint {
				# Deprecated
				# add default buttons & actions for save/delete/cancel
				set record [lindex $value 0]
				set new [lindex $value 1]
				set delete_cmd [lindex $value 2]
				set after_cmd [lindex $value 3]
				set command [string toupper $record]

				# save button; check validation
				pc_button $b.bs image_save "_Save" [list maint_save $b $ns $record $new $wname "$after_cmd"]
				pack $b.bs -side $orient -padx $::sleekPadding -pady $::skinnyPadding

				# delete button; only if maintaining an existing record
				if {$new == 0} {
					if {$delete_cmd == ""} {
						set delete_cmd "if \\\{\\\[confirm $wname \\\"Delete?\\\" -icon warning -default no\\\]\\\} \\\"sql::delete $record;close_window $wname\\\""
					}
					pc_button $b.bd image_delete "_Delete" "namespace eval $ns \"$delete_cmd\""
					pack $b.bd -side $orient -padx $::sleekPadding -pady $::skinnyPadding
				}

				# cancel button; no validation
				pc_button $b.bc image_cancel "Cancel" "namespace eval $ns \"close_window $wname\""
				pack $b.bc -side $orient -padx $::sleekPadding -pady $::skinnyPadding
			}
			-progress {
				# Create progress bar on input buttons, initially not visible
				if {$value} {
					progress::create_local $b
				}
			}
			-enquiry_ap -
			-enquiry_ar -
			-enquiry_gl -
			-enquiry_fa -
			-enquiry_in {
				set module [string toupper [string range $option end-1 end]]
				set number_var $value

				set cmd ""
				set module_name [set login::[string tolower $module]_module(ACCOUNT_NAME)]

				pc_button $b.be image_view "$module_name Enquiry" [list enquiry_button $ns $module $number_var $wname]

				pack $b.be -side $orient -padx $::sleekPadding -pady $::skinnyPadding
			}
			default {
				$b configure $option $value
			}
		}
	}

	set system($b,drawn) 1
}

# process the args for an individual button, returns an array in list format
# old format was: ?novalidate? ?image name?
# new format is: -option ?value? -option ?value?...
# valid button options:
# -novalidate					don't validate the parent window before evaluating the button script
# -image image_name				image to display on button
# -trapevents					activate trapper before evaluating button script; disable once complete
# -delayclose					for "-buttons": close the window AFTER evaluating the button script
# -tk_args						get passed to the tk button command
# -instrument					record button press
# -report_info					display report info option on right-click menu for this button
# any unrecognised options will be returned as the tk_names value in a list 
proc process_button_args { button_args } {

	set defaults {novalidate 0 image "" trapevents 1 delayclose 0 tk_args {} instrument 1 report_info ""}

	if { [string index $button_args 0] != "-" } {
		array set options $defaults
		if { [lindex $button_args 0] == "novalidate" } {
			set options(novalidate) 1
		}
		set options(image) [lindex $button_args 1]
		set options(tk_args) [lindex $button_args 2]

	} else {
		command::arguments $button_args options {} $defaults
		foreach {option value} [array get options] {

			# check for any unknown arguments and add them to the list of args to pass to Tk
			switch -- $option {
				novalidate -
				trapevents -
				delayclose -
				instrument -
				image -
				report_info -
				tk_args {
					# nothing to do
				}
				default {
					lappend options(tk_args) $option $value
					array unset options $option
				}
			}
		}
	}

	return [array get options]
}

# record the pressing of a button
proc instrument_button { button_name } {
	global setup

	# ignore close and cancel buttons:
	if { [lsearch {"close" "cancel"} [string tolower $button_name]]>=0 } { return }

	# convert the window name and button name to identifier-style format
	set window_name [lindex [lindex $setup(frame_stack) end] 2]
	set eventName [instrument::get_event_name $window_name]
	append eventName "."
	append eventName [instrument::get_event_name $button_name]

	if {$eventName != "."} {
		::instrument::event_count $eventName
	}
}

proc enquiry_button {ns module account_number_var wname} {
	set account_number [set ::${ns}::${account_number_var}]
	if {$account_number eq "" || $account_number == 0} {
		set module_name [set login::[string tolower $module]_module(ACCOUNT_NAME)]
		namespace eval $ns [list show_msg $wname "No $module_name selected" -icon warning -title "Account Enquiry Error"]
		return
	}

	# load master record
	if {![query::find account MASTER -key {COMPANY_ID MODULE NUMBER} -COMPANY_ID $login::company_id -MODULE $module -NUMBER $account_number]} {
		set module_name [set login::[string tolower $module]_module(ACCOUNT_NAME)]
		namespace eval $ns [list show_msg $wname "Cannot find $module_name" -icon error -title "Account Enquiry Error"]
		return
	}

	# open new part/customer/supplier/gl account/asset enquiry window for the selected account
	set ns "::Unique[incr ::unique]"

	namespace eval $ns [list source "Maintenance/enquiry_[string tolower $module].tcl"]
	namespace eval $ns {
		source Maintenance/balances.tcl
		source Maintenance/history.tcl
		source Maintenance/select.tcl
		source Transactions/enquiry.tcl
	}

	::${ns}::ENQUIRY [array get account]
}

# Insert or save record, close window
# used by -maint option on input_button
proc maint_save { w ns record new wname after_cmd } {
	# save ok? (validate)
	if {![validate_window $w]} { return }

	# save the record nicely
	# detect duplicate errors (or any other kind also)
	set cmd [expr {$new} ? {"insert"} : {"update"}]
	if {[catch {namespace eval $ns "sql::$cmd $record"} msg]} {
		# check for "duplicate" message? override user message
		if {[string first "duplicate" $msg] != -1} {
			set msg "A record already exists with the details you have entered, please review your entries before trying this again."
		}
		namespace eval $ns [list show_msg $wname $msg -icon error -title "Error during Save"]
	} else {
		if {$after_cmd != ""} {
			namespace eval $ns "$after_cmd"
		}
		namespace eval $ns [list close_window $wname]
	}
}

#====================================================================#

# the input_text widget now defaults the following options:
# -height 4
# -width 50
# -wrap word
# -scroll y

proc input_text { wname field title args } {
	global system setup
	set ns [uplevel {namespace current}]

	set f [system_new_window $wname]

	set ff $f.f
	set ft $ff.t
	set system($ft,ns) $ns
	set system($ft,field) $field

	# define all options
	system_set_defaults $ft lines uppercase lowercase required standard {standard_newline 1} {align 1} {autoselect 0} \
		noentry event {scroll y} note note_config var {expand 0} {tab 1} {enter 0} {fill none} help {address 0} justify

	set draw_args [system_set_options $ft $args]

	set command "[list draw_input_text $f $ff $ft $title] $draw_args"
	system_input_widget $ft $ns $command $wname

	return $ft
}

proc draw_input_text { f ff ft title args} {
	global system setup

	set ns $system($ft,ns)
	set field $system($ft,field)

	set pw [pc_label $f $title]
	frame $ff
	set fx $ff.x
	set fy $ff.y

	# create the text widget
	text $ft -wrap word -height 4 -width 50
	if {[info exists $ns\::$field]} {
		$ft insert end [set $ns\::$field]
	}
	set_focus $ft

	# store options and process undefined ones
	foreach {option value} $args {
		$ft configure $option $value
	}

	# -event? generic event handling: -event {event_name command ...}
	if {$system($ft,event) != ""} {
		foreach {event command} $system($ft,event) {
			# execute the specified command when the event occurs
			::event::bind $ft <$event> "+if \{\[[list namespace eval $ns $command]\]\} break"
		}
	}

	# -align 0? this option removes the widget from the packing/alignment list (see proc pack_labels)
	if {!$system($ft,align)} {
		set labelw [lindex $pw 1]
		array unset system $labelw,pack
	}

	# display text -note? notes for text appear below the text entry window
	if {$system($ft,note) != ""} {
		label $ff.text -text $system($ft,note) -justify left -font [setup font_small]
		pack $ff.text -side bottom -anchor nw
	}

	# -note_config?
	if {$system($ft,note_config) != ""} {
		eval "$ff.text configure $system($ft,note_config)"
	}

	# the tab key is set to focus to the next field by default, unless -tab 0 is set
	if {$system($ft,tab)} {
		::event::bind $ft <Tab> "+focus \[tk_focusNext %W\]; break"
		::event::bind $ft <Shift-Tab> "+focus \[tk_focusPrev %W\]; break"
	}

	# if only one line is allowed to be entered, bind the enter key same as the tab
	if {$system($ft,lines) != "" && $system($ft,lines) == 1} {
		::event::bind $ft <Return> "+focus \[tk_focusNext %W\]; break"
	}

	# -standard "code" text selection option?
	if {$system($ft,standard) != ""} {
		# press F1 to select from list of options
		# press F2 to maintain list of options (requires level 1 UT security)
		::event::bind $ft <Control-F1> "+NULL"
		# the -standard_newline option is set to 0 when you don't want blank lines inserted between text messages
		::event::bind $ft <KeyPress-F1> "+standard_text_selection $ft $system($ft,standard_newline)"
		set f2 ""
		if {$login::user(UT_SECURITY)} {
			::event::bind $ft <Control-F2> "+NULL"
			::event::bind $ft <KeyPress-F2> "+standard_text_selection $ft maint"
			set f2 "\nPress F2 to maintain messages."
		}
		# add any help to any existing help message (popup)
		set system($ft,help) [string trim "Press F1 to select message.$f2\n$system($ft,help)" \n]
	}

	# -help text?
	if {$system($ft,help) != ""} {
		set cmd [list popup_help $system($ft,help) $ft]
		# append all help arguments onto the command; help_args might be used for "-time 0" or other, see support.tcl - popup_help
		::event::bind $ft <FocusIn> $cmd
	}

	# -scroll x/y?
	if {[string match -nocase "*x*" $system($ft,scroll)]} {
		$ft configure -xscrollcommand "$fx set"
		scrollbar $fx -command "$ft xview" -orient h
		pack $fx -side bottom -fill x
	}
	if {[string match -nocase "*y*" $system($ft,scroll)]} {
		$ft configure -yscrollcommand "$fy set"
		scrollbar $fy -command "$ft yview"
		pack $fy -side right -fill y
	}

	# autoselect? select the text when the cursor focuses into this field (done for address fields by default, removed this logic; 29/11/2003)
	if {$system($ft,autoselect)} {
		::event::bind $ft <FocusIn> "$ft tag remove sel 1.0 end;$ft tag add sel 1.0 end"
	}

	pack $ft -side left -anchor nw -expand [expr {!$system($ft,address)}] -fill $system($ft,fill)
	pack $ff -fill both -expand $system($ft,expand)

	# display text -note? notes for text appear below the text entry window

	if {$system($ft,address)} {
		set map [get_map_name $ns [get_currency_code $ns]]
		# make the image then display the country/map button
		if {[make_image $map]} {
			button $ff.addr -image image_$map -command [list zoomin_address $ft] -font [setup font_small] -takefocus 0
			pack $ff.addr -side left -padx $::narrowPadding
		}
	}

	if {$system($ft,expand)} {
		pack $ft -fill both
		pack [lindex $pw 0] -expand $system($ft,expand)
	}

#	bind $ft <KeyRelease> "+set_text_field $ft"
#	bind $ft <ButtonRelease> "+set_text_field $ft"

	# rename the $ft command, provide a wrapper which then updates the text variable after the text is modified
	rename $ft $ft\_
	proc $ft {command args} "
		# perform the specified command
		if {\[info commands $ft\_\] == \"\"} return
		set result \[eval uplevel \[list $ft\_ \$command \$args\]\]

		if {\[string equal \$command insert\] || \[string equal \$command delete\]} {
			set_text_field $ft
		}

		return \$result"

	# display text -var?
	if {$system($ft,var) != ""} {
		label $ff.v -textvariable $ns\::$system($ft,var) -justify left -font [setup font_small]
		pack $ff.v -side left
	}

	# pop-up menu accessible via the right-mouse-click
	::event::bind $ft <ButtonPress-3> "application_popup_entry $ft %X %Y TEXT"

	# set write trace to update variable on screen if changed
	set cmd "NULL $ft;update_text $ns\::$field $ft\_ $ff 0"
	trace variable $ns\::$field w $cmd
	::event::bind $ft <Destroy> +[list trace vdelete $ns\::$field w $cmd]

	# -lines? set-up some logic to restrict the entry length
	set lines $system($ft,lines)
	if {$lines != ""} {
		set cmd "\[input_text_lines $ft $lines\]"
		set system($ft,validate) [list $cmd "The maximum number of lines allowed is $lines.\nPlease truncate this entry."]
	}

	set_system $ft noentry

	set system($ft,drawn) 1
}

proc update_text { field window frame resize name1 name2 op } {
	global system
	global setup

	if {[info commands $window] != ""} {
		set text [set $field]

		set ft $frame.t
		
		set state [$window cget -state]
		$window configure -state normal
		$window delete 1.0 end
		$window insert end $text
		$window configure -state $state
		$window tag add "$system($ft,ns).$frame.text" 0.0 end

		if {$system($ft,justify) != ""} {
			$window tag configure "$system($ft,ns).$frame.text" -justify $system($ft,justify)
		}

		if { $resize } {
			# don't resize if width/height options set in system array (specified by caller)		
			if { $system($frame,text_height) == 0 } {
				set height [text_height $text]
				$window configure -height $height
			}
			if { $system($frame,text_width) == 0 } {
				set font [$window cget -font]				
				set width [text_width $text $font]

				# delete...
				# Charlotte: calculate the width for font_small so that it won't resize the entire window according to bigger font size calculation
				# if {$font == [setup font_small]} {
					# set width [expr $width*[lindex [setup font_small] 1]/[lindex $setup(font) 1]]
				# }
		
				$window configure -width $width
			}
		}
	}
}

proc set_text_field { ft } {
	global system
	set ns $system($ft,ns)
	set field $system($ft,field)

	# get the full text & trim off trailling new line characters
	set text [string trimright [$ft get 1.0 end] "\n"]

	# update the field value
	set_no_trace $ns\::$field $text $ft
}

proc input_text_lines { ft lines } {
	global system
	set ns $system($ft,ns)
	set field $system($ft,field)

	set value [string trim [set $ns\::$field]]

	# if text lines more than allowed return 0 (invalid)
	if {[llength [split $value "\n"]] > $lines} {
		return 0
	}

	return 1
}

# Return the current customer, supplier or company currency
proc get_currency_code { ns } {
	set currency ""
	if {[namespace tail $ns] == "AP"} {
		catch {set currency [set ${ns}::master(PRICING_CURRENCY)]}
		if {$currency != ""} { return $currency }
	}
	catch {set currency [set ${ns}::master(BALANCE_CURRENCY)]}
	if {$currency != ""} { return $currency }
	catch {set currency [set ${ns}::customer(BALANCE_CURRENCY)]}
	if {$currency != ""} { return $currency }
	catch {set currency [set ${ns}::company(CURRENCY_CODE)]}
	if {$currency != ""} { return $currency }

	return $login::company(CURRENCY_CODE)
}

# Return map name based on phone no or currency code
proc get_map_name { ns currencyCode } {
	set phone ""
	set fax ""
	set mobile ""
	catch {set phone [set ${ns}::master(PHONE_NO)]}
	catch {set fax [set ${ns}::master(FAX_NO)]}
	catch {set mobile [set ${ns}::master(MOBILE_NO)]}
	if {$phone == ""}  { catch {set phone [set ${ns}::customer(PHONE_NO)]} }
	if {$fax == ""}    { catch {set fax [set ${ns}::customer(FAX_NO)]} }
	if {$mobile == ""} { catch {set mobile [set ${ns}::customer(MOBILE_NO)]} }

	# Get countryCode countryName mapID
	set countryInfo [address::guess_country $phone $fax $mobile $currencyCode]
	if {$countryInfo ne {}} {
		set map [lindex $countryInfo 2]
	} else {
		set map "world"
	}
	return $map
}

#====================================================================#

proc input_address { wname field title args } {
	set cmd "input_text $wname $field \"$title\" -lines 4 -wrap none -scroll {} -address 1 $args"
	uplevel 1 $cmd
}

proc zoomin_address { ft } {
	global system
	global tcl_platform

	set ns $system($ft,ns)
	set field $system($ft,field)

	# get the full address & trim off trailing new line characters
	set address [string trimright [$ft get 1.0 end] "\n"]

	# URL encoding
	regsub -all "\n" $address "+" address
	regsub -all "," $address "" address
	regsub -all " " $address "+" address
	regsub -all "&" $address "+" address
	regsub -all "\\+\\+" $address "+" address

	# link to mapping website (all google)
	switch -- [get_currency_code $ns] {
		"AU"  { set host "maps.google.com.au" }  
		"AUD" { set host "maps.google.com.au" }  
		"CNY" { set host "ditu.google.cn" }
		"EUR" { set host "maps.google.fr" }  
		"GBP" { set host "maps.google.co.uk" }
		"JPY" { set host "maps.google.co.jp" }
		"NZ"  { set host "maps.google.co.nz" }
		"NZD" { set host "maps.google.co.nz" }
		"RMB" { set host "ditu.google.cn" }
		"STG" { set host "maps.google.co.uk" }
		"UKP" { set host "maps.google.co.uk" }
		"YEN" { set host "maps.google.co.jp" }
		default { set host "maps.google.com" }
	}
	set url "http://$host/maps?f=q&hl=en&geocode=&q=$address"

	# On Win95, even if there is a web browser installed, it cannot
	# open an internet address, only a local .html file.

	# Win98 doesn't like "/"
	if {$::tcl_platform(osVersion) eq "4.10"} {
		regsub -all "/" $url "\\\\" url
	}

	# Substitute & with ^&
	# needs this for XP (which is 5.1)
	if {$::tcl_platform(osVersion) ne "4.10"} {
		regsub -all "&" $url "^&" url
	}

	# Open url (in browser)
	if {$tcl_platform(platform) == "unix"} {
		set command [list xdg-open]
	} else {
		set command [list {*}[auto_execok start] {}]
	}
	exec {*}$command $url &
}

#====================================================================#

# this command is used to display labels & text on the window
# wname = window name
# title = label of the text
# args	= options, note: options before the -text option are for the
#	  labeledwidget and after the -text option for the text widget
# -text = text message to display in the label window
# returns the Tk path of the frame containing the text

proc show_text { wname text args } {
	global system
	set f [system_new_window $wname]
	set ft $f.t
	set ns [uplevel {namespace current}]

	set system($ft,ns) $ns
	set system($ft,wname) $wname
	set system($f,wname) $wname
	set system($ft,parent) [system_parent_window $f $ns]

	# define all options
	system_set_defaults $f scroll {expand no} variable {anchor nw} event {padx 4} {pady 2} {text_width 0} {text_height 0} justify
	system_set_defaults $ft justify

	# remove trailing blank lines from the text
	set text [string trimright $text "\n"]

	# store options and get undefined ones
	set draw_args [system_set_options $f $args]
	
	set command "[list draw_text $f $ns $text] $draw_args"
	system_input_widget $f $ns $command $wname

	return $f
}

proc text_height { text } {
	set lines [split $text "\n"]
	set height [llength $lines]
	return $height
}

proc text_width { text font } {
	set lines [split $text "\n"]
	set max_width 0
	foreach line $lines {
		# set line_width [string length $line]
		set line_width [expr {round([layout_text::text_length $line $font])}]
		if { $line_width > $max_width } {
			set max_width $line_width
		}
	}
	
	# set width [expr {$max_width + 5}]
	return $max_width
}

proc draw_text { f ns text args } {
	global system
	global setup
	

	set height [text_height $text]
	if { $system($f,text_height) != 0 } {
		set height $system($f,text_height)
	}

	# set width [text_width $text]
	# if { $system($f,text_width) != 0 } {
		# set width $system($f,text_width)
	# }

	# delete...
	# Charlotte: calculate the width for font_small so that it won't resize the entire window according to bigger font size calculation
	# foreach {option value} $args {
		# if {$option == "-font"} {
			# if {$value == [setup font_small]} {
				# set width [expr $width*[lindex [setup font_small] 1]/[lindex [setup font] 1]]
			# }
		# }
	# }
	
	frame $f -bd 0 -highlightthickness 0

	text $f.t -wrap word -bd 2 -relief flat -height $height -bg [$f cget -background]
	$f.t insert end $text
	$f.t configure -state disabled
	$f.t tag add "$ns.$f.text" 0.0 end

	foreach {option value} $args {
		switch -- $option {
			default {
				$f.t configure $option $value
			}
		}
	}

	# configure width
	set font [$f.t cget -font]
	set width [text_width $text $font]
	if { $system($f,text_width) != 0 } {
		set width $system($f,text_width)
	}
	$f.t configure -width $width

	if {$system($f,justify) != ""} {
		$f.t tag configure "$ns.$f.text" -justify $system($f,justify)
	}

	# -variable? set trace on variable to update the text widget
	if {$system($f,variable) != ""} {
		set cmd "NULL $f.t;update_text $ns\::$system($f,variable) $f.t $f 1"
		trace variable $ns\::$system($f,variable) w $cmd
		::event::bind $f.t <Destroy> +[list trace vdelete $ns\::$system($f,variable) w $cmd]
	}

	# scroll? y only at present
	if {$system($f,scroll) == "y"} {
		$f.t configure -yscrollcommand "$f.y set"
		scrollbar $f.y -command "$f.t yview"
		pack $f.y -side right -fill y
	}

	# -event? generic event handling: -event {event_name command ...}
	if {$system($f,event) != ""} {
		foreach {event command} $system($f,event) {
			# execute the specified command when the event occurs
			::event::bind $f.t <$event> [list namespace eval $ns $command]
		}
	}

	# pop-up menu accessible via the right-mouse-click
	::event::bind $f.t <ButtonPress-3> "application_popup_entry $f.t %X %Y TEXT"

	pack $f.t -anchor $system($f,anchor) -padx 0 -pady 0 -fill both -expand yes

	pack $f -anchor $system($f,anchor) -padx $system($f,padx) -pady $system($f,pady) -fill both -expand $system($f,expand)

	set system($f,drawn) 1
}

# create miscellaneous widgets
# provide pack arguments as value for 
proc create_widget { wname widget args } {
	global system

	initialise_widgets
	set w [system_new_window $wname]
	set ns [uplevel {namespace current}]

	set system($w,wname) $wname
	set system($w,ns) $ns

	system_set_defaults $w {pack {}} {bind {}}
	set draw_args [system_set_options $w $args]

	set command "[list draw_widget $w $widget] $draw_args"
	system_input_widget $w $ns $command

	return $w
}

proc draw_widget { w widget args} {
	global system

	set command "[list $widget $w] $args"
	eval $command

	set bind_args $system($w,bind)
	foreach {cause effect} $bind_args {
		::event::bind $w <$cause> $effect
	}

	set pack_args $system($w,pack)
	set pack_command "[list pack $w] $pack_args"
	eval $pack_command
}

# evaluate the commands that draw the window
proc draw_window_widgets { w {frame ""} } {
	global system setup delayed_window_closes

	# close any windows that are delayed
	close_delayed_windows

	if { [array names system "$w,draw_commands"] == {} } { return }

	set commands_to_evaluate $system($w,draw_commands)
	set unevaluated_commands {}
	# remove the commands in the system array for now, so they don't get evaluated twice
	set system($w,draw_commands) {}

	foreach command_list $commands_to_evaluate {
		set command [lindex $command_list 0]
		if { $frame != "" } {
			set command_frame [lindex $command_list 1]
			# if the widget to be drawn is not on the selected frame (which hasn't been drawn yet), don't draw it
			if { $frame != $command_frame && ![winfo exists $command_frame] } {
				lappend unevaluated_commands $command_list

				continue
			}
		}

		if { [setup log_draw 0] } { set start [clock clicks] }

		# evaluate the command that will draw the widget
		eval $command

		if { [setup log_draw 0] } {
			if { [regexp {^draw_input_button} $command] } { continue }
			set duration [expr {([clock clicks] - $start) / 1000}]
			if { $duration > 100 } { logToFile "$duration ms to evaluate: '$command'" }
		}
	}

	set system($w,draw_commands) $unevaluated_commands
}

proc close_delayed_windows {} {
	global system delayed_window_closes

	foreach delayed_window $delayed_window_closes {
		foreach {ns wname action} $delayed_window {}
		
		set command [list close_window $wname -delayed]
		if { $action != "" } {
			lappend command -action $action
		}

		namespace eval $ns $command
	}
	set delayed_window_closes {}
}

#====================================================================#

# wname = name of the window array - see "create_window"
# Options:
# "-calc"         process the specified command after every key/button
# "-setfocus"     set to 0 to suppress the initial focus (see support/output)
# "-destroy"      proc/script called on destruction of the window ...
# "-packforget"   hide frame and its contents (for subframe, not toplevel)
# "-packremember" undo packforget
# "-redraw"       proc called when window is redrawn (when child window destroyed)
# "-playback"     replay queued events (default true)
# If you use -destroy, -packforget or -packremember, call input_window again to draw the window
proc input_window { wname args } {
	global system setup
	set ns [uplevel {namespace current}]

	# get the parent window (frames may have {child parent} list)
	# check if already destroyed - suppress errors in certain cases
	if {![info exists system($ns,$wname)] && [lindex $args 0] in {"-packforget" "-destroy"}} return

	set w $system($ns,$wname)
	if {[lindex $w 1] != ""} {
		set w [lindex $w 1]
	}

#	event::disable_trapper

	# draw the fields
	if { $setup(delayed_draw) } {

		# if this window/widget has a parent, only draw if the parent is drawn. Otherwise, add to parent's draw list
		if { [info exists system($w,parent)] && $system($w,parent) != $w } {

			set parent $system($w,parent)
			if { $system($parent,drawn) } {
				event::disable_trapper
				draw_window_widgets $parent $w
			} else {
				set draw_command "[list input_window $wname -recursive 1] $args"

				# input_window command has to be evaluated in the correct (caller's) namespace:
				set command "[list namespace eval $ns $draw_command]"
				lappend system($parent,draw_commands) [list $command $w]
				return
			}
		} else {
			event::disable_trapper
			draw_window_widgets $w
		}
	}

	set toplevel [winfo toplevel $w]
	set top [top_window $w]

	# Save stack info, for right-click "Stack Trace"
	set system($w,stack) [get_stack_trace]

	# define all options
	system_set_defaults $w packforget packremember calc destroy {setfocus 1} redraw {playback 1}

	set recursive 0

	# read all options into the usual system($w,*options*) array
	foreach {option value} [system_set_options $w $args] {
		if { $option == "-recursive" && $value != 0 } {
			set recursive 1
			continue
		}
		$w configure $option $value
	}

	# -destroy? simply return now, see "close_window"
	if {$system($w,destroy) != ""} {
		# set alternate array name: destroy_proc (otherwise gets cleared by
		# next call to input_window)
		set system($w,destroy_proc) $system($w,destroy)
		return
	}

	# -packforget? remove the window/frame from the display
	if {$system($w,packforget) != ""} {
		pack forget $w
		return
	}

	# -packremember? put the window/frame back on the display
	if {$system($w,packremember) != ""} {
		pack $w -expand 1 -fill both
		return
	}

	# -redraw? specify command to execute when the screen is redrawn (by closing any child windows), see "select"
	if {$system($w,redraw) != ""} {
		append system($w,redraw_command) "$system($w,redraw);"
	}

	# call the "close_window" procedure when the window is destroyed (via "X")
	if {[winfo class $w] == "Toplevel"} {
		wm protocol $w WM_DELETE_WINDOW [list namespace eval $ns close_window $wname "destroy"]
	}

	# save the "-setfocus" option here
	# we currently only use it for the print dialog and the variables get unset by the "close_window" function
	# after the "update" command included below (problems working with this on 17/05/01)
	set setfocus $system($w,setfocus)

	set_system $w calc

	pack_labels $w
	update_window $top

	# centre the window for display
	# note: subsequent calls to "input_window" will centre window again
	update idletasks

	# check that the window still exists, we do this here/now because the "update" command above could have
	# triggered an event that closes the window before it's been displayed, eg: user presses Escape
	if {![winfo exists $top]} {
		return
	}

	catch {
		if {$system($w,centre)} {
			if {$system($w,popup)} {
				centre_popup_window $w
			} else {
				centre_window $w
			}
		}
	}

	# display the window (if withdrawn)
	if {[wm state $toplevel] == "withdrawn" && !$recursive} {
		wm deiconify $toplevel
	}

	# focus on the first field in the window, select the contents if active Entry field
	set focus $system($top,focus)
	if {$focus != "" && $setfocus} {
		update
		if {[winfo exists $focus]} {
			if {[lsearch {Frame Labelframe} [winfo class $focus]] != -1} {
				set state "normal"
			} else {
				set state [$focus cget -state]
			}

			focus $focus

			# if the field is not "normal" then focus on the next field
			if {$state != "normal" && !$recursive} {
				# display the window, update then "Tab" to next window
				# don't use the event generate function, this can screw things up with the wrong events

				update

				# re-check the window exists (may not after "update")
				if {[winfo exists $focus]} {
					focus [tk_focusNext $focus]
				}
			}
		}

	}

	if { $recursive } { 
		return 
	}

	# check the window still exists in the system array (it may not due to the 'update' above)
	# playback the queued events
	if { [info exists system($w,playback)] && $system($w,playback) == 1 } {
		after idle [list event::playback $w]
	}
}

# Return execution stack as list of command & argument
proc get_stack_trace {} {
	# Show variables on the stack
	#for {set i [expr [info level] -2]} {$i > 0} {incr i -1} {
	#	set varList [uplevel #$i {info vars}]
	#	logToFile "varList $i $varList"
	#}

	set callStack {}
	set level [info level]
	for {set i 2} {$i < $level} {incr i} {
		set call [uplevel $i {info level 0}]
		# Get fully-qualified command name
		set relative [lindex $call 0]
		if {$relative != ""} {
			set absolute [uplevel $i [list namespace which -command $relative]]
			set arguments [lrange $call 1 end]
			lappend callStack [list $absolute $arguments]
		}
	}
	return $callStack
}

#====================================================================#

# Find max width of all labels in the window
# Get width in pixels, then convert to chars
# This is current used to make the width the same, for all entry fields on a window
# For a notebook, this should be per-tab, not per entire notebook ...
proc pack_labels { w } {
	global system
	set maxwidth 1

	set windows [array names system $w.*,pack]

	# process each window, if it's a label then check width
	foreach name $windows {
		set l [lindex [split $name ","] 0]

		# won't work properly if label title contains tabs or new lines...
		# alternative is to use [winfo reqwidth $l],
		# but this overestimates by quite a bit how much space is required
		set title [$l cget -text]
		set font  [$l cget -font]
		set width [font measure $font -displayof $w $title]
		# set width [winfo $l reqwidth]
		# logToFile "required width of label $l is $width ($title)"

		if {$width > $maxwidth} {
			set maxwidth $width
		}
	}

	foreach name $windows {
		set l [lindex [split $name ","] 0]

		set font  [$l cget -font]
		set charWidth [font measure $font -displayof $w "0"]
		set width [expr {int($maxwidth/$charWidth) + 1}]
		# logToFile "required character width for label $l is $width"

		$l configure -width $width
	}
}

#====================================================================#

# this procedure is used to determine if a window exists

proc window_exists { wname } {
	global system
	set ns [uplevel {namespace current}]

	# get the parent window (frames may have {child parent} list)
	if {[array names system $ns,$wname] != ""} {
		return 1
	}

	return 0
}

#====================================================================#

# return true if the window has been closed:
proc window_closed { wname } {
	set ns [uplevel {namespace current}]
	return [expr {![info exists ::system($ns,$wname)]}]
}

# Close a window - remove displayed window and destroy window object
# Usually called when saving an object, eg. after maintaining the object
# close_window supports two formats of passing arguments
# Historic code uses the old format, the new format is preferred

# args -- old format:
# wname - window name
# second optional parameter, 'action'
# can either be destroy or empty string
#
# New format:
# wname - window name
# Options:
# -action, same valid values as above
# -delay, 1 or 0 (if no value is specified, default is 1)
# if delay = 1, close_window will be delayed until either a new window is drawn, or another existing window is closed

proc close_window { wname args } {
	global system delayed_window_closes
	set ns [uplevel {namespace current}]

	set action ""
	set delay 0
	set delayed 0

	if { [string index $args 0] != "-"} {
		set action [lindex $args 0]
	} else {
		foreach option_value [split $args -] {
			if { $option_value == "" } { continue }
			set option [lindex $option_value 0]
			set value [lindex $option_value 1]

			switch -- $option {
				action {
					set action $value
				}
				delay - delayed {
					if { $value != 0 } {
						set $option 1
					}
				}
				default {
					error "Invalid option \"$option\", value = \"$value\""
				}
			}
		}
	}

	# get the parent window (frames may have {child parent} list)
	if {[catch { set w $system($ns,$wname) } ]} {
		# failure to close window, caused by window problem, opening multiple windows of same name?
		set focus [focus]
		if {$focus == ""} { set focus . }
		set top [winfo toplevel $focus]
		set title [wm title $top]
		log "System error on \"close_window $wname $action\" $title."
		return
	}

	# check if we delay the closing of this window (to prevent the user from interacting with the window below)
	if { [setup delayed_draw 0] && $delay } {
		
		lappend delayed_window_closes [list $ns $wname $action]

		# activate a trapper, we don't want events being processed on this window:
		if { !$event::limbo } { ::event::activate_namespace_trapper $wname $ns }
		return
	}

	# now close any windows that have been delayed
	if { !$delayed } {
		close_delayed_windows
	}

	::event::disable_trapper -window $w
	::event::clear_queue

	if {[lindex $w 1] != ""} {
		set w [lindex $w 1]
	}

	if {[info commands $w] != ""} {
		# does the window have a -destroy procedure?
		if {$action == "destroy"} {
			if {![catch { set destroy_proc $system($w,destroy_proc) }]} {
				if {$destroy_proc != ""} {
					namespace eval $ns $destroy_proc
					return
				}
			}
		}

		# does the parent window have a "redraw_command"?
		if {![catch { set pname $system($w,pname) }]} {
			if {$pname != ""} {
				if {![catch { set pw $system($ns,$pname) }]} {
					# check that the parent window still exists
					# get the redraw command, does it have a value?
					if {![catch { set cmd $system($pw,redraw_command) }]} {
						if {$cmd != ""} {
							after idle [list namespace eval $ns $cmd]
						}
					}

					# perform update on all entry fields, done for goods receipts noentry fields
					update_window $pw
				}
			}
		}

		destroy $w
	}

	return 1
}

#====================================================================#

# this procedure builds a list of procedures/expressions call constantly
# to determine if the various buttons are active or in-active

proc validate_window { w } {
	global system

	set toplevel [winfo toplevel $w]
	set topw [top_window $w]

	set activated_trapper 0
	if { !$event::limbo } {
		event::activate_trapper $topw
		set activated_trapper 1
	}

	set ns $system($topw,ns)
	namespace eval $ns "variable validate_message \"\""

	set valid_field 1
	
	# get a list of all window names that need to be checked
	set names {}
	foreach option {range warning_range high low required valiDATE validate blankok} {

		foreach window_name [array names system $topw*,$option] {
			set name [lindex [split $window_name ,] 0]
			lappend names $name
		}
	}

	# make list of all widgets for the window we are validating, that have one or more check-options
	set names [lsort -unique $names]
	
	# loop through the windows (ie. widgets) and validate all options
	foreach name $names {

		# get some useful fields from the system/window array
		foreach f {ns field title date} {
			if {[catch { set sub_$f $system($name,$f) }]} {
				set sub_$f ""
			}
		}
		
		# set default options
		array set options {range "" warning_range "" high "" low "" required "" valiDATE "" validate "" blankok 0}
		
		# set the options array with the system variables in values
		foreach option {range warning_range high low required valiDATE validate blankok} {
			if {[catch { set value $system($name,$option) }]} { continue }
			set options($option) $value
		}
		
		# check if a required value is set
		if { $options(required) != "" } {
			if {[string trim [set $sub_ns\::$sub_field]] == ""} {
				set required [namespace eval $ns "expr \{$options(required)\}"]
				if {$required} {
					set $ns\::validate_message "An Entry is Required"
					set valid_field 0
					break
				}
			}
		}
		
		# process the date range and warning date
		if {$options(range) != "" || $options(warning_range) != "" } {
			
			# get the value the field is set to
			set sub_value [set $sub_ns\::$sub_field]
			
			# get the dates
			set range_low [lindex $options(range) 0]
			set range_high [lindex $options(range) 1]
			set warning_low [lindex $options(warning_range) 0]
			set warning_high [lindex $options(warning_range) 1]
			
			# do some checks
			if {$sub_value > $range_high && $range_high != ""} {
				set $ns\::validate_message "The entered date is too far in advance, latest date can be [date::display $range_high]"
				set valid_field 0
				break
			} elseif {$sub_value < $range_low && $sub_value != "" && $range_low != ""} {
				set $ns\::validate_message "The entered date is too far behind, earliest date can be [date::display $range_low]"
				set valid_field 0
				break
			} elseif {$sub_value > $warning_high && $warning_high != ""} {
				set response [confirm . "The entered date ([date::display $sub_value]) is not in the normal range (later than [date::display $warning_high]).\nIs this date correct?"]
				if {!$response} {
					set valid_field 0
					break
				}
			} elseif {$sub_value < $warning_low && $sub_value != "" && $warning_low != ""} {
				set response [confirm . "The entered date ([date::display $sub_value]) is not within than the normal range (later than [date::display $warning_low]).\nIs this date correct?"]
				if {!$response} {
					set valid_field 0
					break
				}
			}
		}
		
		# run the valiDATE procdure and display the message if required
		if { $options(valiDATE) != "" } {
			set proc [lindex $options(valiDATE) 0]
			set $ns\::validate_message [lindex $options(valiDATE) 1]
			
			# call the validation procedure
			# return 1=ok, 0=error, -1=error/message already displayed
			set valid [namespace eval $ns set __temp $proc]
			if {!$valid == 1} {
				set valid_field 0
				break
			}
		}
		
		if { $options(validate) != "" } {
			set proc [lindex $options(validate) 0]
			set $ns\::validate_message [lindex $options(validate) 1]

			# call the validation procedure
			# return 1=ok, 0=error, -1=error/message already displayed
			set valid [namespace eval $ns set __temp $proc]
			if {!$valid} {
				set valid_field 0
				break
			}
		}
		
		# check if a numeric value is to high
		if { $options(high) != "" } {
			# get the high test number and the entered value
			catch { set high [namespace eval $ns "expr \{$options(high)\}"] }
			set sub_value [set $sub_ns\::$sub_field]
			if {$sub_value > $high} {
				if {$sub_date == "1"} { set high [date::display $high] }
				set $ns\::validate_message "Value is too high!\nMaximum high = $high"
				set valid_field 0
				break
			}
		}
		
		# check if an numeric value is too low
		if { $options(low) != "" } {
			catch { set low [namespace eval $ns "expr \{$options(low)\}"] }
			set sub_value [set $sub_ns\::$sub_field]
			if {$options(blankok) && $sub_value eq ""} {
				# field is allowed to be blank
			} elseif {$sub_value < $low && $low != ""} {
				if {$sub_date == "1"} { set low [date::display $low] }
				set $ns\::validate_message "Value is too low!\nMinimum low = $low"
				set valid_field 0
				break
			}
		}

		# exit the loop if there is a problem
		if {!$valid_field} { break }
	}

	if { $activated_trapper } {
		event::disable_trapper
	}

	# if the field is invalid then make sure it is visible and place the cursor in it...
	if {!$valid_field} {
		# ensure the window is visible to the user
		check_window_visible $name

		# highlight the entry window with a red background
		# also set the disabledbackground colour (in case the widget is somehow disabled, catch errors, not all widgets support this)
		# name is the tk name of the entry widget (not the list, if any)

		set old_colour [$name cget -background]
		set old_bg ""
		catch { set old_bg [$name cget -disabledbackground] }
		$name configure -background red
		catch { $name configure -disabledbackground red }

		# is this is a combobox widget and the entry field is not displayed
		set combo ""
		set old_combo_colour ""
		set old_combo_bg ""
		if {[info exists system($name,ff)] && [info exists system($name,nameonly)] && $system($name,nameonly) == 1} {
			set combo $system($name,ff)
			set old_combo_colour [$combo cget -background]
			catch { set old_combo_bg [$combo cget -disabledbackground] }
			$combo configure -background red
			catch { $combo configure -disabledbackground red }
		}

		# display the message if there is still one to be displayed
		focus $name
		if {[set $ns\::validate_message] != ""} {
			fb_messageBox -title "Data Validation Error" -message [set $ns\::validate_message] -icon error -parent $toplevel
			# re-focus the window (sometimes the messageBox routine loses it again)
			catch { focus $name }
		}

		# restore normal colours
		catch { $name configure -background $old_colour }
		catch { $name configure -disabledbackground $old_bg }

		if {$combo != ""} {
			catch { $combo configure -background $old_combo_colour }
			catch { $combo configure -disabledbackground $old_combo_bg }
		}
		return 0
	}

	return 1
}

#====================================================================#

proc check_window_visible { window } {
	# first select the window contents (called from <FocusIn> event)
	catch {$window selection range 0 end}

	# check the window path name, if any parent is a tabbed notebook then we must ensure
	# that the correct tab page is made visible, eg: window = .window3.tab3.f0.f2.e1
	# the "tab" represents a tab page, the ".f0" afterwards represents page 0

	set wlist [split $window "."]
	set w ""
	set body_bit 0
	foreach bit $wlist {
		if {$body_bit} {
			# found a tab window, raise the page
			set i [string range $bit 1 end]
			Rnotebook:raise $w $i
			break
		}
		# build up the window name, eg: .window1.tab3
		if {$bit == "body"} {
			set body_bit 1
		} elseif {$bit != ""} {
			append w ".$bit"
		}
	}
}

#====================================================================#

proc update_window { w } {
	global system system_calc system_noentry

	# get the top window, this represents the visible window
	set w [top_window $w]

	# process all calculations from input_char/input_window
	# calculations may be stored against the top window, or any of its children
	# previously this code used to just use [array names system_calc $w*] to match all of these widgets
	# but this doesn't work, e.g when your widget is .window1 (i.e the main window), will also match .window100, .window101 etc
	set calc_widgets [concat [array names system_calc $w] [array names system_calc "${w}.*"]]
	foreach win [lsort $calc_widgets] {
		# check for errors, the array might not exist after evaluating some commands as they may close an open window
		if {![catch { set value $system_calc($win) }]} {
			set ns $system($win,ns)
			namespace eval $ns $value
		}
	}

	# process all noentry options
	set noentry_widgets [concat [array names system_noentry $w] [array names system_noentry "${w}.*"]]
	foreach win [lsort [array names system_noentry $w*]] {
		# does the window still exist?
		if {[info commands $win] == "" || ![info exists system_noentry($win)]} { continue }

		set value $system_noentry($win)
		set ns $system($win,ns)
		set field ""
		catch {set field $system($win,field)}

		# store original background colour (if not done already)
		if {[catch { set system($win,normal_bg) } ]} {
			set system($win,normal_bg) [$win cget -background]
		}

		# calculate the -noentry value
		if {[catch { set noentry [namespace eval $ns "expr \{$value\}"] }]} { continue }

		# test ...
		# make sure the noentry value can be treated as a boolean
		# string is boolean doesn't return true for e.g 2
		# use an if statement to test, since it's an if statement we want it to work for
		if {[catch { if {$noentry} {} }]} {
			logToFile "Invalid noentry expression '$value' - gives result '$noentry'"
			continue
		}

		# check again the window still exists after evaluating the expression
		if {[info commands $win] == ""} { continue }

		# configure the input window
		# if "field" is blank then assume lookup name field
		if {$noentry} {
			if {$field != ""} {
				# notes: the foreground colour gets overwritten when a field is disabled, look at options database
				if {[$win cget -state] != "disabled"} {
					$win configure -state "disabled" -background lightgray
				}
			} else {
				if {[$win cget -background] != "lightgray"} {
					$win configure -state "disabled" -background lightgray
				}
			}
		} else {
			if {$field != ""} {
				if {[$win cget -state] != "normal"} {
					# set state then background (two separate commands)
					$win configure -state normal
					$win configure -background $system($win,normal_bg)
				}
			} else {
				if {[$win cget -background] != $system($win,normal_bg)} {
					# set state then background (two separate commands)
					$win configure -state normal
					$win configure -background $system($win,normal_bg)
				}
			}
		}
	}

	return {}
}

# Set widget option in the system array
proc set_system { w option } {
	global system system_calc system_noentry

	set value $system($w,$option)

	# don't bother storing values if "0" only
	# if the option contains a value then store in the more specific
	# system array, either system_calc() or system_noentry()
	if {$value != "" && $value != "0"} {
		set system_$option\($w) $value
	}
}

#====================================================================#

proc plural { number } {
	if {$number == 1} { return "" }
	return "s"
}

